<?php

/**
 * Shopgate GmbH
 *
 * URHEBERRECHTSHINWEIS
 *
 * Dieses Plugin ist urheberrechtlich geschützt. Es darf ausschließlich von Kunden der Shopgate GmbH
 * zum Zwecke der eigenen Kommunikation zwischen dem IT-System des Kunden mit dem IT-System der
 * Shopgate GmbH über www.shopgate.com verwendet werden. Eine darüber hinausgehende Vervielfältigung, Verbreitung,
 * öffentliche Zugänglichmachung, Bearbeitung oder Weitergabe an Dritte ist nur mit unserer vorherigen
 * schriftlichen Zustimmung zulässig. Die Regelungen der §§ 69 d Abs. 2, 3 und 69 e UrhG bleiben hiervon unberührt.
 *
 * COPYRIGHT NOTICE
 *
 * This plugin is the subject of copyright protection. It is only for the use of Shopgate GmbH customers,
 * for the purpose of facilitating communication between the IT system of the customer and the IT system
 * of Shopgate GmbH via www.shopgate.com. Any reproduction, dissemination, public propagation, processing or
 * transfer to third parties is only permitted where we previously consented thereto in writing. The provisions
 * of paragraph 69 d, sub-paragraphs 2, 3 and paragraph 69, sub-paragraph e of the German Copyright Act shall remain unaffected.
 *
 * @author Shopgate GmbH <interfaces@shopgate.com>
 */
$file = rtrim(DIR_FS_CATALOG, "/") . "/" . rtrim(DIR_WS_CLASSES, "/")
	. "/discount_coupon.php";
if (file_exists($file)) {
	include_once($file);
}

class ShopgateCouponModel
{
	function __construct(){}
	
	/**
	 * @param $shopCoupon
	 *
	 * @return bool
	 */
	public function isCouponInsideOfATimeFrame($shopCoupon)
	{
		$now   = strtotime(date("Y-m-d H:i:s"));
		$begin = strtotime($shopCoupon["coupons_date_start"]);
		$end   = strtotime($shopCoupon["coupons_date_end"]);

		return ($now < $begin && $now > $end) ? false : true;
	}

	/**
	 * @param $pId
	 * @param $code
	 *
	 * @return bool
	 */
	public function isProductExcludedFromCategory($pId, $code)
	{
		$categoryQuery = "SELECT count(*) AS cnt 
							FROM " . TABLE_DISCOUNT_COUPONS . " AS dc 
								JOIN " . TABLE_DISCOUNT_COUPONS_TO_CATEGORIES
			. " AS dctca ON dctca.coupons_id = dc.coupons_id 
								JOIN " . TABLE_CATEGORIES . " AS c ON c.categories_id = dctca.categories_id
								JOIN " . TABLE_PRODUCTS_TO_CATEGORIES . " AS ptc ON ptc.categories_id = c.categories_id
								JOIN " . TABLE_PRODUCTS . " AS p ON p.products_id = ptc.products_id 
							WHERE dc.coupons_id = \"{$code}\" AND dc.coupons_max_use > 0 AND p.products_id = {$pId};";
		$result        = ShopgateWrapper::db_query($categoryQuery);
		$categoryCount = ShopgateWrapper::db_fetch_array($result);

		return (!empty($categoryCount) && !empty($categoryCount["cnt"])
			&& $categoryCount["cnt"] > 0) ? true : false;
	}

	/**
	 * @param $pId
	 * @param $code
	 *
	 * @return bool
	 */
	public function isProductExcluded($pId, $code)
	{
		$productQuery = "SELECT count(*) AS cnt 
								FROM " . TABLE_DISCOUNT_COUPONS . " AS dc 
									LEFT JOIN "
			. TABLE_DISCOUNT_COUPONS_TO_PRODUCTS . " AS dctpr ON dctpr.coupons_id = dc.coupons_id
									LEFT JOIN " . TABLE_PRODUCTS . " AS p ON p.products_id = dctpr.products_id
								WHERE dc.coupons_id = \"{$code}\" AND dc.coupons_max_use > 0 AND p.products_id = {$pId};";
		$result       = ShopgateWrapper::db_query($productQuery);
		$productCount = ShopgateWrapper::db_fetch_array($result);

		return (!empty($productCount) && !empty($productCount["cnt"])
			&& $productCount["cnt"] > 0) ? true : false;
	}

	/**
	 * @param $pId
	 * @param $code
	 *
	 * @return bool
	 */
	public function isManufacturerExcluded($pId, $code)
	{
		$manufacturersQuery = "SELECT count(*) AS cnt 
								FROM " . TABLE_DISCOUNT_COUPONS . " AS dc 
									JOIN "
			. TABLE_DISCOUNT_COUPONS_TO_MANUFACTURERS . " AS dctm ON dctm.coupons_id = dc.coupons_id
									JOIN " . TABLE_MANUFACTURERS . " AS m ON m.manufacturers_id = dctm.manufacturers_id
									JOIN " . TABLE_PRODUCTS . " AS p ON p.manufacturers_id = m.manufacturers_id
								WHERE dc.coupons_id = \"{$code}\" AND dc.coupons_max_use > 0 AND p.products_id = {$pId};";
		$result             = ShopgateWrapper::db_query($manufacturersQuery);
		$manufacturerCount  = ShopgateWrapper::db_fetch_array($result);

		return (!empty($manufacturerCount) && !empty($manufacturerCount["cnt"])
			&& $manufacturerCount["cnt"] > 0) ? true : false;
	}

	/**
	 * @param $code
	 * @param $customerId
	 *
	 * @return bool
	 */
	public function isCustomerExcluded($code, $customerId)
	{
		$customerQuery = "SELECT count(*) AS cnt 
								FROM " . TABLE_DISCOUNT_COUPONS . " AS dc 
									LEFT JOIN "
			. TABLE_DISCOUNT_COUPONS_TO_CUSTOMERS . " AS dctc ON dctc.coupons_id = dc.coupons_id
									LEFT JOIN " . TABLE_CUSTOMERS . " AS c ON c.customers_id = dctc.customers_id
								WHERE dc.coupons_id = \"{$code}\" AND dc.coupons_max_use > 0 AND c.customers_id = {$customerId};";

		$result        = ShopgateWrapper::db_query($customerQuery);
		$customerCount = ShopgateWrapper::db_fetch_array($result);

		return (!empty($customerCount) && !empty($customerCount["cnt"])
			&& $customerCount["cnt"] > 0) ? true : false;
	}

	/**
	 * @param $code
	 * @param $customerId
	 *
	 * @return bool
	 */
	public function customerHasOrders($code, $customerId)
	{
		$orderQuery = "SELECT count(*) AS cnt 
								FROM " . TABLE_DISCOUNT_COUPONS . " AS dc 
									LEFT JOIN "
			. TABLE_DISCOUNT_COUPONS_TO_ORDERS . " AS dcto ON dcto.coupons_id = dc.coupons_id
									LEFT JOIN " . TABLE_ORDERS . " AS o ON o.orders_id = dcto.orders_id
								WHERE dc.coupons_id = \"{$code}\" AND dc.coupons_max_use > 0 AND o.customers_id = {$customerId};";

		$result     = ShopgateWrapper::db_query($orderQuery);
		$orderCount = ShopgateWrapper::db_fetch_array($result);

		return (!empty($orderCount) && !empty($orderCount["cnt"])
			&& $orderCount["cnt"] > 0) ? true : false;
	}

	/*#######################
	 # 	redeem_coupons
	 ######################*/

	/**
	 * @param $code
	 *
	 * @return ressource
	 */
	public function redeemValidCouponByCode($code)
	{
		$query = "UPDATE discount_coupons AS dc 
						SET dc.coupons_number_available = dc.coupons_number_available - 1 
				  	WHERE dc.coupons_id = \"{$code}\" 
				  		AND dc.coupons_number_available > 0;";

		return ShopgateWrapper::db_query($query);
	}
} 
