<?php

/**
 * Shopgate GmbH
 *
 * URHEBERRECHTSHINWEIS
 *
 * Dieses Plugin ist urheberrechtlich geschützt. Es darf ausschließlich von Kunden der Shopgate GmbH
 * zum Zwecke der eigenen Kommunikation zwischen dem IT-System des Kunden mit dem IT-System der
 * Shopgate GmbH über www.shopgate.com verwendet werden. Eine darüber hinausgehende Vervielfältigung, Verbreitung,
 * öffentliche Zugänglichmachung, Bearbeitung oder Weitergabe an Dritte ist nur mit unserer vorherigen
 * schriftlichen Zustimmung zulässig. Die Regelungen der §§ 69 d Abs. 2, 3 und 69 e UrhG bleiben hiervon unberührt.
 *
 * COPYRIGHT NOTICE
 *
 * This plugin is the subject of copyright protection. It is only for the use of Shopgate GmbH customers,
 * for the purpose of facilitating communication between the IT system of the customer and the IT system
 * of Shopgate GmbH via www.shopgate.com. Any reproduction, dissemination, public propagation, processing or
 * transfer to third parties is only permitted where we previously consented thereto in writing. The provisions
 * of paragraph 69 d, sub-paragraphs 2, 3 and paragraph 69, sub-paragraph e of the German Copyright Act shall remain unaffected.
 *
 * @author Shopgate GmbH <interfaces@shopgate.com>
 */
class ShopgateCustomerModel
{
    function __construct() { }

    /**
     * @param ShopgateCart $cart
     *
     * @return array
     */
    public function getAddress(ShopgateCart $cart)
    {
        $deliveryAddress = $cart->getDeliveryAddress();
        $address         = empty($deliveryAddress) ? $cart->getInvoiceAddress()
            : $deliveryAddress;
        $delivery        = array();
        if (!empty($address)) {
            $delivery['zone_id']    = $this->getStateId(
                ShopgateMapper::getShoppingsystemStateCode($address->getState())
            );
            $delivery['country_id'] =
                $this->getCountryId($address->getCountry());
        }

        return $delivery;
    }

    /**
     * @param $currentState
     *
     * @return mixed
     */
    private function getStateId($currentState)
    {
        $qry         = "SELECT z.zone_id FROM " . TABLE_ZONES
                       . " AS z where z.zone_code = \"{$currentState}\" ;";
        $result      = ShopgateWrapper::db_query_print_on_err($qry);
        $resultArray = ShopgateWrapper::db_fetch_array($result);

        return $resultArray['zone_id'];
    }

    /**
     * @param $country
     *
     * @return mixed
     */
    private function getCountryId($country)
    {
        $qry         = "SELECT c.countries_id FROM " . TABLE_COUNTRIES
                       . " AS c where c.countries_iso_code_2 = \"{$country}\" ;";
        $result      = ShopgateWrapper::db_query_print_on_err($qry);
        $resultArray = ShopgateWrapper::db_fetch_array($result);

        return $resultArray['countries_id'];
    }

    /**
     * Gets token or creates one if needed
     *
     * @param ShopgateCustomer $customer
     * @return null|string
     */
    public function getTokenForCustomer(ShopgateCustomer $customer)
    {
        $token = $this->getToken($customer);
        if (!$token) {
            $id = $this->getIdFromEmail($customer->getMail());
            if ($id) {
                $token = $this->createToken($customer);
                $customer->setCustomerToken($token);
                $this->saveToken($customer);
            }
        }
        return $token;
    }

    /**
     * @param ShopgateCustomer $customer
     * @return string|null
     */
    private function getToken(ShopgateCustomer $customer)
    {
        $query       = "SELECT c.customer_token FROM " . TABLE_CUSTOMERS_SHOPGATE_CUSTOMER
                       . " AS c where c.customer_id = \"{$customer->getCustomerId()}\" ;";
        $result      = ShopgateWrapper::db_query_print_on_err($query);
        $resultArray = ShopgateWrapper::db_fetch_array($result);

        return isset($resultArray['customer_token']) ? $resultArray['customer_token'] : null;
    }

    /**
     * Simple ID retrieval
     *
     * @param $email - customer's email address
     * @return int|null
     */
    private function getIdFromEmail($email)
    {
        $query       = "SELECT c.customers_id FROM " . TABLE_CUSTOMERS
                       . " AS c where c.customers_email_address = \"{$email}\" ;";
        $result      = ShopgateWrapper::db_query_print_on_err($query);
        $resultArray = ShopgateWrapper::db_fetch_array($result);

        return isset($resultArray['customers_id']) ? (int)$resultArray['customers_id'] : null;
    }

    /**
     * Returns customer token based on
     *
     * @param ShopgateCustomer $customer
     * @return string
     */
    private function createToken(ShopgateCustomer $customer)
    {
		return md5($customer->getMail() . $customer->getCustomerId());
    }

    /**
     * @param ShopgateCustomer $customer
     * @return bool|mysqli_result|resource
     */
    private function saveToken(ShopgateCustomer $customer)
    {
        $data['customer_id']    = $customer->getCustomerId();
        $data['customer_token'] = $customer->getCustomerToken();
        return ShopgateWrapper::db_execute_query(TABLE_CUSTOMERS_SHOPGATE_CUSTOMER, $data);
    }

    /**
     * @param $token
     * @return int|null
     */
    public function getIdFromToken($token)
    {
        $resultArray = ShopgateWrapper::db_select_query(
            TABLE_CUSTOMERS_SHOPGATE_CUSTOMER,
            array('customer_token' => $token),
            array('customer_id'));
        
        $resultArray = array_pop($resultArray);
        return isset($resultArray['customer_id']) ? (int) $resultArray['customer_id'] : null;
    }
} 
