<?php
/**
 * Shopgate GmbH
 *
 * URHEBERRECHTSHINWEIS
 *
 * Dieses Plugin ist urheberrechtlich geschützt. Es darf ausschließlich von Kunden der Shopgate GmbH
 * zum Zwecke der eigenen Kommunikation zwischen dem IT-System des Kunden mit dem IT-System der
 * Shopgate GmbH über www.shopgate.com verwendet werden. Eine darüber hinausgehende Vervielfältigung, Verbreitung,
 * öffentliche Zugänglichmachung, Bearbeitung oder Weitergabe an Dritte ist nur mit unserer vorherigen
 * schriftlichen Zustimmung zulässig. Die Regelungen der §§ 69 d Abs. 2, 3 und 69 e UrhG bleiben hiervon unberührt.
 *
 * COPYRIGHT NOTICE
 *
 * This plugin is the subject of copyright protection. It is only for the use of Shopgate GmbH customers,
 * for the purpose of facilitating communication between the IT system of the customer and the IT system
 * of Shopgate GmbH via www.shopgate.com. Any reproduction, dissemination, public propagation, processing or
 * transfer to third parties is only permitted where we previously consented thereto in writing. The provisions
 * of paragraph 69 d, sub-paragraphs 2, 3 and paragraph 69, sub-paragraph e of the German Copyright Act shall remain unaffected.
 *
 * @author Shopgate GmbH <interfaces@shopgate.com>
 */

/**
 * @author Konstantin Kiritsenko <konstantin@kiritsenko.com>
 */
class ShopgateOrderModel
{
	/**
	 * @param $customerToken
	 * @param $customerLanguage
	 * @param $limit
	 * @param $offset
	 * @param $orderDateFrom
	 * @param $sortOrder
	 * @return array
	 * @throws ShopgateLibraryException
	 */
	public function getCustomerOrders($customerToken, $customerLanguage, $limit, $offset, $orderDateFrom, $sortOrder)
	{
		$response      = array();
		$customerModel = new ShopgateCustomerModel();
		$customerId    = $customerModel->getIdFromToken($customerToken);

		if (!is_numeric($customerId)) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_CUSTOMER_TOKEN_INVALID,
				"Shopgate Plugin - no customer found with token: " . $customerToken,
				true
			);
		}
		$data = array(
			'limit'     => $limit,
			'offset'    => $offset,
			'date_from' => $orderDateFrom,
			'sort'      => $sortOrder
		);

		$orders = $this->_getOrdersByCustomer($customerId, $data);

		foreach ($orders as $row) {

			$shopgateOrder = $this->_getShopgateOrder($row['orders_id']);

			$shopgateExternalOrder = new ShopgateExternalOrder();
			$shopgateExternalOrder->setOrderNumber($shopgateOrder['shopgate_order_number']);
			$shopgateExternalOrder->setExternalOrderId($row['orders_id']);
			$shopgateExternalOrder->setExternalOrderNumber($row['orders_id']);
			$shopgateExternalOrder->setStatusName($this->_getStatusName($row['orders_status'], $customerLanguage));
			$shopgateExternalOrder->setCreatedTime(date(DateTime::ISO8601, strtotime($row['date_purchased'])));
			$shopgateExternalOrder->setMail($row['customers_email_address']);
			$shopgateExternalOrder->setPhone($row['customers_telephone']);
			$shopgateExternalOrder->setInvoiceAddress($this->_getOrderAddress($row, 'billing'));
			$shopgateExternalOrder->setDeliveryAddress($this->_getOrderAddress($row, 'delivery'));
			$shopgateExternalOrder->setCurrency($row['currency']);
			$shopgateExternalOrder->setAmountComplete($row['order_total'] ? $row['order_total'] : null);
			$shopgateExternalOrder->setPaymentMethod($row['payment_method']);
			$shopgateExternalOrder->setIsPaid(($shopgateOrder['is_paid']) ? $shopgateOrder['is_paid'] : null);
			$shopgateExternalOrder->setPaymentTransactionNumber($row['orders_total_id']);
			$itemsAndTaxes = $this->_getOrderItemsFormatted($row['orders_id'], $row['currency']);
			$shopgateExternalOrder->setOrderTaxes($itemsAndTaxes[1]);
			$shopgateExternalOrder->setExtraCosts($this->_getExtraCosts($row['orders_id']));
			$shopgateExternalOrder->setItems($itemsAndTaxes[0]);

			array_push($response, $shopgateExternalOrder);
		}

		return $response;
	}

	/**
	 * Gets Shopgate order from system order ID
	 *
	 * @param $sysOrderId
	 * @return mixed[]
	 * @throws ShopgateLibraryException
	 */
	private function _getShopgateOrder($sysOrderId)
	{
		$orders = ShopgateWrapper::db_select_query(
			TABLE_ORDERS_SHOPGATE_ORDER,
			array('orders_id' => $sysOrderId)
		);

		return array_pop($orders);
	}

	/**
	 * Gets address data
	 *
	 * @param $order - orders table row
	 * @param $type  - type of column, either "billing" or "delivery"
	 * @return mixed[]
	 */
	private function _getOrderAddress($order, $type)
	{
		$type .= '_';
		$name            = explode(' ', $order[$type . 'name']);
		$shopgateAddress = new ShopgateAddress();
		$shopgateAddress->setFirstName($name[0]);
		$shopgateAddress->setLastName($name[1]);
		$shopgateAddress->setCompany($order[$type . 'company']);
		$shopgateAddress->setPhone($order['customers_telephone']);
		$shopgateAddress->setStreet1($order[$type . 'street_address']);
		$shopgateAddress->setStreet2($order[$type . 'suburb']);
		$shopgateAddress->setCity($order[$type . 'city']);
		$shopgateAddress->setZipcode($order[$type . 'postcode']);
		$shopgateAddress->setState($order[$type . 'state']);
		$countryCodeRows = ShopgateWrapper::db_select_query(TABLE_COUNTRIES, array('countries_name' => $order[$type . 'country']));
		$shopgateAddress->setCountry($countryCodeRows[0]['countries_iso_code_2']);

		return $shopgateAddress->toArray();
	}

	/**
	 * Gets customer data based on data provided
	 *
	 * @param       $customerId
	 * @param array $data (
	 *                    'limit'     => ...,
	 *                    'offset'    => ...,
	 *                    'date_from' => ...,
	 *                    'sort'      => ...
	 *                    )
	 * @return array
	 * @throws ShopgateLibraryException
	 */
	private function _getOrdersByCustomer($customerId, $data = array())
	{
		/**
		 * Query DB for order data
		 */
		$sqlQuery = "SELECT c.*, ot.orders_total_id, ot.value AS order_total "
			. "FROM " . TABLE_ORDERS . " AS c "
			. "LEFT JOIN " . TABLE_ORDERS_TOTAL . " AS ot "
			. "ON c.orders_id = ot.orders_id "
			. "WHERE " . "c.customers_id = '{$customerId}' "
			. "AND ot.class='ot_total'";

		if (!empty($data['date_from'])) {
			$sqlQuery .= " AND c.date_purchased >= '" . $data['date_from'] . "'";
		}

		if (!empty($data['sort'])) {
			$sortOrder = strtoupper(str_replace('created_', '', $data['sort']));
			$sqlQuery .= " ORDER BY c.date_purchased {$sortOrder}";
		}

		if (!empty($data['limit'])) {
			$sqlQuery .= " LIMIT " . $data['limit'];
			if (isset($data['offset'])) {
				$sqlQuery .= " OFFSET " . $data['offset'];
			}
		}

		$sqlQuery .= ";";

		$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);

		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error selecting an order.", true
			);
		}

		/**
		 * Queue orders up
		 */
		$orders = array();
		while ($row = ShopgateWrapper::db_fetch_array($queryResult)) {
			array_push($orders, $row);
		}

		return $orders;
	}

	/**
	 * @param $sysOrderId
	 * @param $currency
	 * @return array
	 */
	private function _getOrderItemsFormatted($sysOrderId, $currency)
	{
		$results = ShopgateWrapper::db_select_query(TABLE_ORDERS_PRODUCTS, array('orders_id' => $sysOrderId));
		$orderItems = array();
		$orderTaxes = array();
		foreach ($results as $orderProduct) {
			$item         = new ShopgateExternalOrderItem;

			$item->setItemNumber($orderProduct["products_id"]);
			$item->setItemNumberPublic($orderProduct["products_model"]);
			$item->setQuantity($orderProduct["products_quantity"]);
			$item->setName($orderProduct["products_name"]);
			$item->setUnitAmount($orderProduct["products_price"]);
			$gross = $orderProduct["products_price"] * (1 + $orderProduct["products_tax"] / 100);
			$item->setUnitAmountWithTax($gross);
			$item->setTaxPercent($orderProduct["products_tax"]);
			$item->setCurrency($currency);

			$orderTaxAdded = false;
			foreach ($orderTaxes as $orderTax) {
				if ($orderTax->getTaxPercent() == $orderProduct["products_tax"]) {
					$orderTax->setAmount($orderTax->getAmount() + $gross * $orderProduct["products_quantity"]);
					$orderTaxAdded = true;
				}
			}
			if (!$orderTaxAdded) {
				$orderTax = new ShopgateExternalOrderTax();
				$orderTax->setAmount($gross * $orderProduct["products_quantity"]);
				$orderTax->setTaxPercent($orderProduct["products_tax"]);
				$orderTax->setLabel(round($orderProduct["products_tax"], 2) . '%');
				$orderTaxes[] = $orderTax;
			}


			$orderItems[] = $item;
		}

		return array($orderItems, $orderTaxes);
	}

	/**
	 * @param $status
	 * @param $customerLanguage
	 * @return mixed
	 */
	protected function _getStatusName($status, $customerLanguage)
	{
		$customerLanguageCode = explode("_", $customerLanguage);
		$customerLanguageCode = $customerLanguageCode[0];
		$customerLanguageCodeRow = ShopgateWrapper::db_select_query(TABLE_LANGUAGES, array('code' => $customerLanguageCode));
		$customerLanguageCodeRow = $customerLanguageCodeRow[0];
		if (empty($customerLanguageCodeRow)) {
			$customerLanguageCodeRow['languages_id'] = 1;
		}
		$name = ShopgateWrapper::db_select_query(TABLE_ORDERS_STATUS, array('orders_status_id' => $status, 'language_id' => $customerLanguageCodeRow['languages_id']));

		return $name[0]['orders_status_name'];
	}

	/**
	 * @param $sysOrderId
	 * @return array
	 */
	protected function _getExtraCosts($sysOrderId)
	{
		$totals = ShopgateWrapper::db_select_query(TABLE_ORDERS_TOTAL, array('orders_id'=> $sysOrderId));
		$extraCosts = array();
		foreach ($totals as $total) {
			$extraCost = new ShopgateExternalOrderExtraCost();
			$type = '';
			switch ($total['class']) {
				case 'ot_subtotal':
					break;
				case 'ot_total' :
					break;
				case 'ot_tax' :
					break;
				case 'ot_shipping' :
					$type = 'shipping';
					break;
				default:
					$type = 'misc';
			}
			if (empty($type)) {
				continue;
			}
			$extraCost->setType($type);
			$extraCost->setAmount($total['value']);
			$extraCost->setLabel($total['title']);
			$extraCosts[] = $extraCost;
		}
		return $extraCosts;
	}
}