<?php
/**
 * Shopgate GmbH
 *
 * URHEBERRECHTSHINWEIS
 *
 * Dieses Plugin ist urheberrechtlich geschützt. Es darf ausschließlich von Kunden der Shopgate GmbH
 * zum Zwecke der eigenen Kommunikation zwischen dem IT-System des Kunden mit dem IT-System der
 * Shopgate GmbH über www.shopgate.com verwendet werden. Eine darüber hinausgehende Vervielfältigung, Verbreitung,
 * öffentliche Zugänglichmachung, Bearbeitung oder Weitergabe an Dritte ist nur mit unserer vorherigen
 * schriftlichen Zustimmung zulässig. Die Regelungen der §§ 69 d Abs. 2, 3 und 69 e UrhG bleiben hiervon unberührt.
 *
 * COPYRIGHT NOTICE
 *
 * This plugin is the subject of copyright protection. It is only for the use of Shopgate GmbH customers,
 * for the purpose of facilitating communication between the IT system of the customer and the IT system
 * of Shopgate GmbH via www.shopgate.com. Any reproduction, dissemination, public propagation, processing or
 * transfer to third parties is only permitted where we previously consented thereto in writing. The provisions
 * of paragraph 69 d, sub-paragraphs 2, 3 and paragraph 69, sub-paragraph e of the German Copyright Act shall remain unaffected.
 *
 * @author Shopgate GmbH <interfaces@shopgate.com>
 */
require_once(dirname(__FILE__) . '/shopgate_library/shopgate.php');
require_once(DIR_FS_CATALOG . 'ext/modules/shopgate/base/shopgate_config.php');
require_once(DIR_FS_CATALOG . 'ext/modules/shopgate/model/ShopgateModelLoader.php');
define('SHOPGATE_PLUGIN_VERSION', '2.9.15');


/**
 * osCommerce-Plugin for Shopgate
 */
class ShopgatePluginOsCommerce extends ShopgatePlugin
{
	/**
	 * @var ShopgateConfigOsCommerce
	 */
	protected $config;
	/**
	 * @var int
	 */
	protected $taxZone;
	/**
	 * @var string
	 */
	protected $swisscartVersion;
	/**
	 * @var int
	 */
	protected $swisscartImageCount;
	/**
	 * @var string
	 */
	protected $language;
	/**
	 * @var string
	 */
	protected $currency;
	/**
	 * @var int
	 */
	protected $itemHighlightOrderIndex;
	/**
	 * @var int
	 */
	protected $langID;
	/**
	 * @var ShopgateCartItemModel $cartItemModel
	 */
	protected $cartItemModel;
	/**
	 * @var ShopgateItemModel $itemModel
	 */
	protected $itemModel;
	/**
	 * @var ShopgateCouponModel $couponModel
	 */
	protected $couponModel;
	/**
	 * @var ShopgateLocationModel $locationModel
	 */
	protected $locationModel;
	/**
	 * @var ShopgateTaxModel $taxModel
	 */
	protected $taxModel;
	/**
	 * @var ShopgateReviewModel $reviewModel
	 */
	protected $reviewModel;	
	/**
	 * @var ShopgateCategoryModel $categoryModel
	 */
	protected $categoryModel;

	/**
	 * @var ShopgateCustomerModel $customerModel
	 */
	protected $customerModel;
	
	/** @var  ShopgateOrderModel $orderModel */
	protected $orderModel;
	
	/**
	 * @return bool
	 * @throws ShopgateLibraryException
	 */
	public function startup()
	{
		// Check Shoppingsystem-Type
		$this->swisscartVersion = defined("SWISSCART_VERSION") ? SWISSCART_VERSION : '';
		$this->config           = new ShopgateConfigOsCommerce();

		if (!empty($_REQUEST['shop_number'])) {
			$this->config->loadByShopNumber($_REQUEST['shop_number']);
		}

		// Map config language code to OSC language ID
		$langCode   = $this->config->getLanguage();
		$sql        = "SELECT languages_id, directory FROM " . TABLE_LANGUAGES . " WHERE code = '$langCode'";
		$langResult = ShopgateWrapper::db_query($sql);
		if (!$langResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error selecting language.", true
			);
		}
		$language = ShopgateWrapper::db_fetch_array($langResult);
		if (empty($language) || empty($language['languages_id'])
			|| empty($language['directory'])
		) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - language code [$langCode] does not exist.",
				true
			);
		}
		$this->langID       = $language['languages_id'];
		$this->language     = $language["directory"];
		$this->currency     = $this->config->getCurrency();
		$sql                = "SELECT value FROM " . TABLE_CURRENCIES . " WHERE code = '" . $this->currency . "'";
		$result             = ShopgateWrapper::db_query($sql);
		$_currency          = ShopgateWrapper::db_fetch_array($result);
		$this->exchangeRate = $_currency["value"];
		$this->taxZone      = $this->config->getTaxZoneId();

		if (!defined("DIR_FS_LANGUAGES")) {
			define('DIR_FS_LANGUAGES', dirname(__FILE__) . '/base/admin/includes/languages/');
		}

		$langFiles = array($this->language . '.php');

		foreach ($langFiles as $langFile) {
			include_once(DIR_FS_LANGUAGES . "/$langFile");
		}

		$properties     = get_object_vars($this);
		$classWhitelist = array('ShopgateReviewXmlModel', 'ShopgateCategoryXmlModel', 'ShopgateItemXmlModel');
		foreach ($properties as $className => $propertyValue) {
			if (substr($className, -strlen("Model")) == "Model") {
				$classWhitelist[] = "Shopgate" . ucfirst($className);
				$classWhitelist[] = "AbstractShopgate" . ucfirst($className);
			}
		}

		$modelLoader = new ShopgateModelLoader($classWhitelist);
		$modelLoader->includeModels();
		$includeList = $modelLoader->getIncludedFileList();
		$includeList[] = dirname(__FILE__) . '/base/includes/shopgate_currencies.php';

		// instantiate models dynamically
		foreach ($includeList as $class => $path) {
			include_once $path;
			if (class_exists($class)) {
				if (strpos($class, "Abstract") !== 0) {
					$variableName    = $modClassName = str_replace("Shopgate", "", $class);
					$variableName{0} = strtolower($modClassName{0});
					$instance        = new ReflectionClass($class);
					if ($instance->getConstructor()) {
						$cTorParams = $instance->getConstructor()->getParameters();
					} else {
						$cTorParams = array();
					}
					$argList = array();

					foreach ($cTorParams AS $param) {
						$argList[] = $this->{$param->getName()};
					}
					$this->{$variableName} = $instance->newInstanceArgs($argList);
				}
			}
		}

		return true;
	}

	/**
	 * @param string           $user
	 * @param string           $pass
	 * @param ShopgateCustomer $customer
	 *
	 * @throws ShopgateLibraryException
	 */
	public function registerCustomer($user, $pass, ShopgateCustomer $customer)
	{
		require_once(rtrim(DIR_FS_CATALOG, "/") . "/" . rtrim(DIR_WS_FUNCTIONS, "/") . '/password_funcs.php');
		/** @var ShopgateCustomer $customer */
		$customer        = $customer->utf8Decode($this->config->getEncoding());
		$user            = $this->stringFromUtf8($user, $this->config->getEncoding());
		$userExistResult = tep_db_query(
			"SELECT count(1) AS exist FROM " . TABLE_CUSTOMERS . " AS c WHERE c.customers_email_address = \"{$user}\";"
		);
		$userCount       = tep_db_fetch_array($userExistResult);
		$userCount       = $userCount['exist'];
		$encPass         = tep_encrypt_password($pass);
		$date            = date("Y-m-d H:i:s");

		if ((int)$userCount >= 1) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::REGISTER_USER_ALREADY_EXISTS, '', true
			);
		}

		if (!defined(TABLE_CUSTOMERS_MEMO)) {
			define(TABLE_CUSTOMERS_MEMO, "customers_memo");
		}

		if (!defined(TABLE_CUSTOMERS_INFO)) {
			define(TABLE_CUSTOMERS_INFO, "customers_info");
		}

		$customerData = array(
			'customers_firstname'     => $customer->getFirstName(),
			'customers_lastname'      => $customer->getLastName(),
			'customers_email_address' => $customer->getMail(),
			'customers_telephone'     => $customer->getPhone(),
			'customers_newsletter'    => 0,
			"customers_gender"        => $customer->getGender(),
			'customers_password'      => $encPass
		);

		tep_db_perform(TABLE_CUSTOMERS, $customerData);
		$userId = tep_db_insert_id();

		$customersInfo = array(
			'customers_info_id'                         => $userId,
			'customers_info_number_of_logons'           => 0,
			'customers_info_date_account_created'       => $date,
			'customers_info_date_account_last_modified' => $date
		);
		tep_db_perform(TABLE_CUSTOMERS_INFO, $customersInfo);
		
		$customer->setCustomerId($userId);
		$this->customerModel->getTokenForCustomer($customer);
		
		$addresses   = array();
		$addressList = $customer->getAddresses();

		foreach ($addressList as $key => $address) {
			foreach ($addressList as $secondKey => $secondAddress) {
				if ($address->equals($secondAddress) && $secondKey < $key) {
					$key = $secondKey;
				}
			}
			$addresses[$key] = $address;
		}

		$defaultAddress = true;
		foreach ($addresses as $address) {
			/** @var ShopgateAddress $address */
			$stateCode    = ShopgateMapper::getShoppingsystemStateCode($address->getState());
			$zoneQuery    = tep_db_query(
				"SELECT z.zone_id,z.zone_name from " . TABLE_ZONES . " as z where z.zone_code = '" . $stateCode . "'"
			);
			$zoneResult   = tep_db_fetch_array($zoneQuery);
			$countryQuery = tep_db_query(
				"SELECT c.countries_id from " . TABLE_COUNTRIES . " as c where c.countries_iso_code_2 ='"
				. $address->getCountry() . "'"
			);

			$countryResult = tep_db_fetch_array($countryQuery);
			$addressData   = array(
				"customers_id"         => $userId,
				"entry_company"        => $address->getCompany(),
				"entry_zone_id"        => $zoneResult['zone_id'],
				"entry_country_id"     => $countryResult['countries_id'],
				"entry_firstname"      => $address->getFirstName(), "entry_lastname" => $address->getLastName(),
				"entry_gender"         => $address->getGender(),
				"entry_street_address" => $address->getStreet1(),
				"entry_postcode"       => $address->getZipcode(),
				"entry_city"           => $address->getCity(),
				"entry_state"          => $zoneResult['zone_name']
			);

			tep_db_perform(TABLE_ADDRESS_BOOK, $addressData);
			if ($defaultAddress) {
				$addressId = tep_db_insert_id();
				$query     = "UPDATE " . TABLE_CUSTOMERS
							 . " as c SET customers_default_address_id = {$addressId} WHERE c.customers_id={$userId}";
				tep_db_query($query);
				$defaultAddress = false;
			}
		}
	}

	############################################################################
	## SHOPGATE-CONNECT                                                       ##
	############################################################################

	/**
	 * @param string $user
	 * @param string $pass
	 *
	 * @return ShopgateCustomer
	 * @throws ShopgateLibraryException
	 */
	public function getCustomer($user, $pass)
	{
		// Get the customers data first
		$sqlQuery = "SELECT\n" . "\tcustomer.customers_id,\n"
					. "\tcustomer.customers_gender,\n"
					. "\tcustomer.customers_firstname,\n"
					. "\tcustomer.customers_lastname,\n" . "\tcustomer.customers_dob,\n"
					. "\tcustomer.customers_telephone,\n"
					. "\tcustomer.customers_password,\n"
					. "\tcustomer.customers_email_address,\n"
					. "\tcustomer.customers_default_address_id,\n"
					. "\tcustomer.customers_newsletter\n" . "FROM\n" . "\t"
					. TABLE_CUSTOMERS . " AS customer\n" . "WHERE\n"
					. "\tcustomer.customers_email_address = '"
					. ShopgateWrapper::db_input($user) . "'\n" . ";";

		// Get results
		$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);
		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error selecting the customer.", true
			);
		}

		// Get all customer data and all addresses out of the address
		$customerData = ShopgateWrapper::db_fetch_array($queryResult);

		// Check if a user was found
		if (empty($customerData)) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_WRONG_USERNAME_OR_PASSWORD,
				'User: ' . $user
			);
		}

		// Check if the user has been authenticated correctly
		if (!ShopgateWrapper::validate_password(
			$pass, $customerData['customers_password']
		)
		) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_WRONG_USERNAME_OR_PASSWORD,
				'User: ' . $user
			);
		}

		// Get all user-addresses
		$customerAddresses = array();
		$sqlQuery          = "SELECT\n" . "\taddress.address_book_id,\n"
							 . "\taddress.entry_gender,\n" . "\taddress.entry_company,\n"
							 . "\taddress.entry_firstname,\n" . "\taddress.entry_lastname,\n"
							 . "\taddress.entry_street_address,\n"
							 . "\taddress.entry_postcode,\n" . "\taddress.entry_city,\n"
							 . "\tcountry.countries_iso_code_2 AS entry_country,\n"
							 . "\tzone.zone_code AS entry_state_zone\n" . "FROM\n" . "\t"
							 . TABLE_ADDRESS_BOOK . " AS address\n" . "\t\tLEFT JOIN "
							 . TABLE_COUNTRIES
							 . " AS country ON (country.countries_id = address.entry_country_id)\n"
							 . "\t\tLEFT JOIN " . TABLE_ZONES
							 . " AS zone ON (zone.zone_id = address.entry_zone_id)\n" . "WHERE\n"
							 . "\taddress.customers_id = " . $customerData['customers_id'] . "\n"
							 . ";";

		// Get results
		$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);
		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error selecting the customer addresses.",
				true
			);
		}

		// Get all customer data and all addresses out of the customers address
		$i = 0;
		while ($customerAddress = ShopgateWrapper::db_fetch_array($queryResult)) {
			// Put the default address at the top of the list
			if ($customerAddress['address_book_id'] == $customerData['customers_default_address_id']) {
				$customerAddresses[0] = $customerAddress;
				// Sort instantly before more items are inserted. This can happen only once, because we select by a unique id
				ksort($customerAddresses);
			} else {
				$customerAddresses[++$i] = $customerAddress;
			}
		}

		// Create list of shopgate addresses
		$addresses = array();
		foreach ($customerAddresses as $customerAddress) {
			// Get the correct state
			try {
				$stateCode = ShopgateMapper::getShopgateStateCode(
					$customerAddress["entry_country"],
					$customerAddress["entry_state_zone"]
				);
			} catch (ShopgateLibraryException $e) {
				// Happens if the state code can not be mapped to ISO use system state code
				$stateCode = $customerAddress['entry_state_zone'];
			}

			// Invoice and delivery addresses are always identical in this shoppingsystem
			$addressType = ShopgateAddress::BOTH;
			$addressGender
						 = $customerAddress["entry_gender"] == "m" ? ShopgateAddress::MALE
				: ShopgateAddress::FEMALE;

			// Now fill a new address data container
			$address = new ShopgateAddress();
			$address->setId($customerAddress["address_book_id"]);
			$address->setAddressType($addressType);
			$address->setGender($addressGender);
			$address->setFirstName($customerAddress["entry_firstname"]);
			$address->setLastName($customerAddress["entry_lastname"]);
			$address->setBirthday(null);
			$address->setCompany($customerAddress["entry_company"]);
			$address->setStreet1($customerAddress["entry_street_address"]);
			$address->setStreet2(null);
			$address->setZipcode($customerAddress["entry_postcode"]);
			$address->setCity($customerAddress["entry_city"]);
			$address->setCountry($customerAddress["entry_country"]);
			$address->setState($stateCode);
			$address->setPhone($customerData["customers_telephone"]);
			$address->setMobile("");
			$address->setMail($customerData["customers_email_address"]);

			// Add it to the addresses list
			$addresses[] = $address;
		}

		// Set the customer data
		$customer = new ShopgateCustomer();
		$customer->setCustomerId($customerData["customers_id"]);
		$customer->setCustomerNumber($customerData["customers_id"]);
		$customer->setCustomerGroup(null);
		$customer->setCustomerGroupId(null);
		$customer->setGender(
			$customerData["customers_gender"] == "m" ? ShopgateCustomer::MALE : ShopgateCustomer::FEMALE
		);
		$customer->setFirstName($customerData["customers_firstname"]);
		$customer->setLastName($customerData["customers_lastname"]);
		$customer->setBirthday(
			!empty($customerData["customers_dob"]) ? strftime("%Y-%m-%d", $customerData["customers_dob"]) : null
		);
		$customer->setPhone($customerData["customers_telephone"]);
		$customer->setMobile("");
		$customer->setMail($customerData["customers_email_address"]);
		$customer->setNewsletterSubscription($customerData["customers_email_address"] == 1 ? 1 : 0);

		$customer->setAddresses($addresses);
		$customer->setCustomerToken($this->customerModel->getTokenForCustomer($customer));

		return $customer;
	}

	############################################################################
	## ADD ORDER                                                              ##
	############################################################################

	/**
	 * @param ShopgateOrder $oShopgateOrder
	 *
	 * @return array
	 * @throws ShopgateLibraryException
	 */
	public function addOrder(ShopgateOrder $oShopgateOrder)
	{
		// #1: Check if the order already exists (already imported)
		// -> TABLE_ORDERS_SHOPGATE_ORDER must exist to be able to execute queries on it!
		if (!$this->tableExists(TABLE_ORDERS_SHOPGATE_ORDER)) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error: \"" . TABLE_ORDERS_SHOPGATE_ORDER
				. "\" table is missing.", true
			);
		}
		$sqlQuery = "SELECT\n" . "\t`shopgateorder`.`shopgate_order_id`,\n"
					. "\t`shopgateorder`.`orders_id`,\n"
					. "\t`shopgateorder`.`shopgate_order_number`\n" . "FROM\n" . "\t`"
					. TABLE_ORDERS_SHOPGATE_ORDER . "` AS `shopgateorder`\n" . "WHERE\n"
					. "\t`shopgateorder`.`shopgate_order_number` = '"
					. $oShopgateOrder->getOrderNumber() . "'\n" . ";";

		// -> Get results
		$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);
		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error selecting shopgate orders.", true
			);
		}

		// -> Prevent an existing order to be inserted twice (or more)
		$order = ShopgateWrapper::db_fetch_array($queryResult);
		if (!empty($order)) {
			throw new ShopgateLibraryException(ShopgateLibraryException::PLUGIN_DUPLICATE_ORDER);
		}

		// Get currency for the order
		$currencyCode = strtoupper($oShopgateOrder->getCurrency());
		// Load from currencies table
		$sqlQuery    = "SELECT * FROM " . TABLE_CURRENCIES . " WHERE UPPER(code) = '"
					   . ShopgateWrapper::db_input($currencyCode) . "' LIMIT 1";
		$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);
		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error on selecting the currency with the currency code \"$currencyCode\".",
				true
			);
		}
		$currency = ShopgateWrapper::db_fetch_array($queryResult);
		if (!$currency) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - No currency found for the currency code \"$currencyCode\".",
				true
			);
		}

		// Create a complete order
		$statusHistoryComment     = '';
		$isShippingBlockedComment = '';
		$ordersStatusId           = 1;
		$order                    = $this->saveNewOrder(
			$oShopgateOrder, $currency, $ordersStatusId, $statusHistoryComment, $isShippingBlockedComment
		);

		// -> Save as Shopgate-order
		$this->saveShopgateOrder($oShopgateOrder, $order['orders_id']);

		// -> Save all products that have been ordered (including their specific attributes)
		$errors         = array();
		$ordersProducts = $this->saveOrdersItems($oShopgateOrder, $order['orders_id'], $currency, $errors);

		// -> Save the total amount
		$orderTotals = $this->saveOrdersTotal($oShopgateOrder, $order['orders_id'], $currency, $ordersProducts);

		// -> Save status history and corresponding comments
		$this->saveOrdersStatusHistory(
			$oShopgateOrder, $order['orders_id'],
			$order['payment_method'] == ENTRY_PAYMENT_SHOPGATE_GENERIC ? 1 : 0,
			$ordersStatusId, $statusHistoryComment, $isShippingBlockedComment,
			$errors
		);
		if ($this->config->getSendOrderConfirmationMail()) {
			$this->sendOrderConfirmationMail($order, $ordersProducts, $orderTotals, $statusHistoryComment);
		}

		return array(
			'external_order_id'     => $order['orders_id'],
			'external_order_number' => $order['orders_id']
		);
	}

	/**
	 * @param $order
	 * @param $products
	 * @param $orderTotals
	 * @param $statusHistoryComment
	 */
	public function sendOrderConfirmationMail($order, $products, $orderTotals, $statusHistoryComment)
	{
		
		$currencies     = new shopgate_currencies();
		$ordersProducts = "";
		
		foreach ($products as $product) {
			if(method_exists($currencies, "calculate_price")){
				$displayPrice = $currencies->calculate_price(
					$product['final_price'], $product['products_tax'], $product['products_quantity']
				);
			}else{
				$displayPrice = ($product["products_price"] * (1+($product['products_tax']/100))) * $product["products_quantity"];
			}
			$displayPrice = $currencies->format($displayPrice, 0, $order['currency']);
			$ordersProducts .= $product['products_quantity'] . ' x ' . $product['products_name'] . ' ('
							   . $product['products_model'] . ') = ' . $displayPrice . "\n";
		}
		$emailOrder = STORE_NAME . "\n" .
					  EMAIL_SEPARATOR . "\n" .
					  EMAIL_TEXT_ORDER_NUMBER . ' ' . $order['orders_id'] . "\n" .
					  EMAIL_TEXT_INVOICE_URL . ' ' . tep_href_link(
						  FILENAME_ACCOUNT_HISTORY_INFO, 'order_id=' . $order['orders_id'], 'SSL', false
					  ) . "\n" .
					  EMAIL_TEXT_DATE_ORDERED . ' ' . strftime(DATE_FORMAT_LONG) . "\n\n";
		if (!empty($statusHistoryComment)) {
			$emailOrder .= tep_db_output($statusHistoryComment) . "\n\n";
		}
		$emailOrder .= EMAIL_TEXT_PRODUCTS . "\n" .
					   EMAIL_SEPARATOR . "\n" .
					   $ordersProducts .
					   EMAIL_SEPARATOR . "\n";

		foreach ($orderTotals as $orderTotal) {
			$emailOrder .= strip_tags($orderTotal['title']) . ' ' . strip_tags($orderTotal['text']) . "\n";
		}

		if ($order->content_type != 'virtual') {
			$emailOrder .= "\n" . EMAIL_TEXT_DELIVERY_ADDRESS . "\n" .
						   EMAIL_SEPARATOR . "\n" ;
			$deliveryAddress = array(
				'firstname' => $order['delivery_firstname'],
				'lastname' => $order['delivery_lastname'],
				'company' => $order['delivery_company'],
				'street_address' => $order['delivery_street_address'],
				'suburb' => $order['delivery_suburb'],
				'city' => $order['delivery_city'],
				'postcode' => $order['delivery_postcode'],
				'state' => $order['delivery_state'],
				'zone_id' => $order['delivery_zone_id'],
				'country_id' => $order['delivery_country_id'],
			);
			$emailOrder .= tep_address_format($order['customers_address_format_id'], $deliveryAddress, false, '', "\n") . "\n\n";
		}

		$billingAddress = array(
			'firstname' => $order['billing_firstname'],
			'lastname' => $order['billing_lastname'],
			'company' => $order['billing_company'],
			'street_address' => $order['billing_street_address'],
			'suburb' => $order['billing_suburb'],
			'city' => $order['billing_city'],
			'postcode' => $order['billing_postcode'],
			'state' => $order['billing_state'],
			'zone_id' => $order['billing_zone_id'],
			'country_id' => $order['billing_country_id'],
		);
		$emailOrder .= "\n" . EMAIL_TEXT_BILLING_ADDRESS . "\n" .
					   EMAIL_SEPARATOR . "\n";
		$emailOrder .= tep_address_format($order['customers_address_format_id'], $billingAddress, false, '', "\n") . "\n\n";
		$emailOrder .= EMAIL_TEXT_PAYMENT_METHOD . "\n" .
					   EMAIL_SEPARATOR . "\n";

		$emailOrder .= $order['payment_method'] . "\n\n";
		tep_mail(
			$order['customers_name'], $order['customers_email_address'],
			EMAIL_TEXT_SUBJECT, $emailOrder, STORE_OWNER, STORE_OWNER_EMAIL_ADDRESS
		);
	}

	/**
	 * @param ShopgateOrder $order
	 *
	 * @return array
	 * @throws ShopgateLibraryException
	 */
	public function updateOrder(ShopgateOrder $order)
	{
		// save UTF-8 payment infos (to build proper json)
		$paymentInfosUtf8 = $order->getPaymentInfos();
		// data needs to be utf-8 decoded for äöüß and the like to be saved correctly
		/** @var ShopgateOrder $order */
		$order = $order->utf8Decode($this->config->getEncoding());

		$qry
			= "
			SELECT
				o.*,
				so.shopgate_order_id,
				so.shopgate_order_number,
				so.is_paid,
				so.is_shipping_blocked,
				so.payment_infos
			FROM " . TABLE_ORDERS . " o
			INNER JOIN " . TABLE_ORDERS_SHOPGATE_ORDER . " so ON (so.orders_id = o.orders_id)
			WHERE so.shopgate_order_number = '{$order->getOrderNumber()}'
		";

		$result  = ShopgateWrapper::db_query_print_on_err($qry);
		$dbOrder = ShopgateWrapper::db_fetch_array($result);

		if ($dbOrder == false) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_ORDER_NOT_FOUND,
				"Shopgate order number: '{$order->getOrderNumber()}'."
			);
		}

		$errorOrderStatusIsSent                     = false;
		$errorOrderStatusAlreadySet                 = array();
		$statusShoppingsystemOrderIsPaid            = $dbOrder['is_paid'];
		$statusShoppingsystemOrderIsShippingBlocked = $dbOrder['is_shipping_blocked'];
		$status                                     = $dbOrder["orders_status"];

		// check if shipping is already done, then throw at end of method a OrderStatusIsSent - Exception
		if ($status == $this->config->getOrderStatusShipped()
			&& ($statusShoppingsystemOrderIsShippingBlocked || $order->getIsShippingBlocked())
		) {
			$errorOrderStatusIsSent = true;
		}

		if ($order->getUpdatePayment() == 1) {
			if (!is_null($statusShoppingsystemOrderIsPaid)
				&& $order->getIsPaid() == $statusShoppingsystemOrderIsPaid
				&& !is_null($dbOrder['payment_infos'])
				&& $dbOrder['payment_infos'] == $this->jsonEncode(
					$paymentInfosUtf8
				)
			) {
				$errorOrderStatusAlreadySet[] = 'payment';
			}

			if (is_null($statusShoppingsystemOrderIsPaid) || $order->getIsPaid() != $statusShoppingsystemOrderIsPaid) {
				$orderStatus = array();
				if ($order->getIsPaid()) {
					$orderStatusComment = ENTRY_NEW_PAYMENT_STATUS_IS_PAID_COMMENT_TEXT;
				} else {
					$orderStatusComment = ENTRY_NEW_PAYMENT_STATUS_IS_NOT_PAID_COMMENT_TEXT;
				};

				$orderStatus['comments'] = $this->stringFromUtf8($orderStatusComment, $this->config->getEncoding());

				$this->saveSingleOrdersStatusHistory(
					$dbOrder['orders_id'], $status, $orderStatusComment,
					date('Y-m-d H:i:s')
				);

				// update the shopgate order status information
				$ordersShopgateOrder
					= array(
					"is_paid"  => (int)$order->getIsPaid(),
					"modified" => "now()",
				);
				ShopgateWrapper::db_execute_query(
					TABLE_ORDERS_SHOPGATE_ORDER, $ordersShopgateOrder, "update",
					"shopgate_order_id = {$dbOrder['shopgate_order_id']}"
				);

				// Save status in order
				$orderData                  = array();
				$orderData["orders_status"] = $status;
				$orderData["last_modified"] = date('Y-m-d H:i:s');
				ShopgateWrapper::db_execute_query(
					TABLE_ORDERS, $orderData, "update",
					"orders_id = {$dbOrder['orders_id']}"
				);
			}

			// update payment infos
			if (!is_null($dbOrder['payment_infos'])
				&& $dbOrder['payment_infos'] != $this->jsonEncode($paymentInfosUtf8)
			) {
				$dbPaymentInfos = $this->jsonDecode($dbOrder['payment_infos'], true);
				$paymentInfos   = $order->getPaymentInfos();
				$histories      = array();

				switch ($order->getPaymentMethod()) {
					case ShopgateOrder::SHOPGATE:
					case ShopgateOrder::INVOICE:
					case ShopgateOrder::COD:
						break;
					case ShopgateOrder::PREPAY:
						if (isset($dbPaymentInfos['purpose'])
							&& $paymentInfos['purpose']
							   != $dbPaymentInfos['purpose']
						) {
							$comments = $this->stringFromUtf8(
								ENTRY_PAYMENT_UPDATED
								. ENTRY_PAYMENT_PREPAY_PAYMENT_PURPOSE . ' '
								. $paymentInfos['purpose'],
								$this->config->getEncoding()
							);

							// Order is not paid yet
							$histories[] = array(
								"orders_id"         => $dbOrder["orders_id"],
								"orders_status_id"  => $status,
								"date_added"        => date('Y-m-d H:i:s'),
								"customer_notified" => false,
								"comments"          => ShopgateWrapper::db_prepare_input($comments)
							);
						}
						break;
					case ShopgateOrder::DEBIT:
						// Save additional information into status history comments
						$newPaymentMethodExtraInfo = '';
						$newPaymentMethodExtraInfo .= ENTRY_PAYMENT_ELV_INFOTEXT
													  . "\n";
						$newPaymentMethodExtraInfo
							.= ENTRY_PAYMENT_ELV_TEXT_BANK_ACCOUNT_OWNER . " "
							   . $paymentInfos["bank_account_holder"] . "\n";
						$newPaymentMethodExtraInfo
							.= ENTRY_PAYMENT_ELV_TEXT_BANK_ACCOUNT_NUMBER . " "
							   . $paymentInfos["bank_account_number"] . "\n";
						$newPaymentMethodExtraInfo
							.= ENTRY_PAYMENT_ELV_TEXT_BANK_CODE . " "
							   . $paymentInfos["bank_code"] . "\n";
						$newPaymentMethodExtraInfo
							.= ENTRY_PAYMENT_ELV_TEXT_BANK_NAME . " "
							   . $paymentInfos["bank_name"] . "\n";

						$comments = $this->stringFromUtf8(
							ENTRY_PAYMENT_UPDATED . $newPaymentMethodExtraInfo,
							$this->config->getEncoding()
						);
						$comments .= $this->_createPaymentInfos(
							$paymentInfos, $dbOrder['orders_id'], $status, false
						);

						$histories[]
							= array(
							"orders_id"         => $dbOrder["orders_id"],
							"orders_status_id"  => $status,
							"date_added"        => date('Y-m-d H:i:s'),
							"customer_notified" => false,
							"comments"          => ShopgateWrapper::db_prepare_input(
								$comments
							)
						);

						break;
					case ShopgateOrder::PAYPAL:
						// Save payment infos in history
						$history             = $this->_createPaymentInfos(
							$paymentInfos, $dbOrder["orders_id"], $status
						);
						$history['comments'] = $this->stringFromUtf8(
								ENTRY_PAYMENT_UPDATED,
								$this->config->getEncoding()
							) . $history['comments'];
						$histories[]         = $history;
						break;
					default:
						// mobile_payment
						// Save payment infos in history
						$history             = $this->_createPaymentInfos(
							$paymentInfos, $dbOrder["orders_id"], $status
						);
						$history['comments'] = $this->stringFromUtf8(
								ENTRY_PAYMENT_UPDATED,
								$this->config->getEncoding()
							) . $history['comments'];
						$histories[]         = $history;
						break;
				}

				foreach ($histories as $history) {
					$this->saveSingleOrdersStatusHistory(
						$history['orders_id'], $history['orders_status_id'],
						$history['comments'], $history['date_added']
					);
				}
			}

			$ordersShopgateOrder = array(
				"payment_infos" => $this->jsonEncode(
					$paymentInfosUtf8
				), "modified"   => "now()"
			);
			ShopgateWrapper::db_execute_query(
				TABLE_ORDERS_SHOPGATE_ORDER, $ordersShopgateOrder, 'update',
				"shopgate_order_id = {$dbOrder['shopgate_order_id']}"
			);
		}

		if ($order->getUpdateShipping() == 1) {
			if (!is_null($statusShoppingsystemOrderIsShippingBlocked)
				&& $order->getIsShippingBlocked()
				   == $statusShoppingsystemOrderIsShippingBlocked
			) {
				$errorOrderStatusAlreadySet[] = 'shipping';
			} else {
				if ($status != $this->config->getOrderStatusShipped()) {
					if ($order->getIsShippingBlocked() == 1) {
						$status
							= $this->config->getOrderStatusShippingBlocked();
					} else {
						$status = $this->config->getOrderStatusOpen();
					}
				}

				$orderStatus                      = array();
				$orderStatus["orders_id"]         = $dbOrder["orders_id"];
				$orderStatus["date_added"]        = date('Y-m-d H:i:s');
				$orderStatus["customer_notified"] = false;
				$orderStatus['orders_status_id']  = $status;
				if ($order->getIsShippingBlocked() == 0) {
					$orderStatus["comments"] = ENTRY_NEW_SHIPPING_STATUS_SHIPPING_APPROVED_COMMENT_TEXT;
				} else {
					$orderStatus['comments'] = ENTRY_NEW_SHIPPING_STATUS_SHIPPING_BLOCKED_COMMENT_TEXT;
				}

				$orderStatus['comments'] = $this->stringFromUtf8(
					$orderStatus['comments'], $this->config->getEncoding()
				);
				$this->saveSingleOrdersStatusHistory(
					$orderStatus['orders_id'], $orderStatus['orders_status_id'],
					$orderStatus['comments'], $orderStatus['date_added']
				);

				$ordersShopgateOrder = array(
					"is_shipping_blocked" => (int)$order->getIsShippingBlocked(), "modified" => "now()"
				);
				ShopgateWrapper::db_execute_query(
					TABLE_ORDERS_SHOPGATE_ORDER, $ordersShopgateOrder, 'update',
					"shopgate_order_id = {$dbOrder['shopgate_order_id']}"
				);
				// Save status in order
				$orderData                  = array();
				$orderData["orders_status"] = $status;
				$orderData["last_modified"] = date('Y-m-d H:i:s');
				ShopgateWrapper::db_execute_query(
					TABLE_ORDERS, $orderData, 'update',
					"orders_id = {$dbOrder['orders_id']}"
				);
			}
		}

		if ($errorOrderStatusIsSent) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_ORDER_STATUS_IS_SENT
			);
		}

		if (!empty($errorOrderStatusAlreadySet)) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_ORDER_ALREADY_UP_TO_DATE,
				implode(',', $errorOrderStatusAlreadySet), true
			);
		}

		return array(
			'external_order_id'     => $dbOrder["orders_id"],
			'external_order_number' => $dbOrder["orders_id"]
		);
	}

	/**
	 * @return array|mixed|mixed[]
	 * @throws Exception
	 * @throws ShopgateLibraryException
	 */
	public function createPluginInfo()
	{
		$aInfos = $this->executeLoaders($this->_getPluginInfoLoaders(), array(), null);

		return $aInfos;
	}

	/**
	 * @return array
	 */
	public function getSettings()
	{
		$taxes                         = array();
		$taxes['product_tax_classes']  = $this->getTaxClasses();
		$taxes['customer_tax_classes'] = array(
			'id'         => 1,
			'key'        => 'default',
			'is_default' => 1
		); // no customer tax classes in osCommerce

		// Tax rates are pretty much a combination of tax rules and tax rates in osCommerce. So we're using them to generate both: 
		$oscTaxRates = $this->getTaxRates();
		$taxRates    = array();
		$taxRules    = array();
		foreach ($oscTaxRates as $oscTaxRate) {
			$description = empty($oscTaxRate['tax_description']) 
				? $oscTaxRate['countries_iso_code_2'] . '-' . $oscTaxRate['tax_rate'] 
				: $oscTaxRate['tax_description'];
			// build and append tax rate
			$taxRates[] = array(
				'id'                        => $oscTaxRate['tax_rates_id'],
				'key'                       => $oscTaxRate['tax_rates_id'],
				'display_name'              => $description,
				'tax_percent'               => $oscTaxRate['tax_rate'],
				'country'                   => $oscTaxRate['countries_iso_code_2'],
				'state'                     => (!empty($oscTaxRate['countries_iso_code_2'])
												&& !empty($oscTaxRate['zone_code']))
					? ShopgateMapper::getShopgateStateCode(
						$oscTaxRate['countries_iso_code_2'],
						$oscTaxRate['zone_code']
					) : '', 'zip_code_type' => 'all',
			);

			// build and append tax rule
			$taxRules[] = array(
				'id'                   => $oscTaxRate['tax_rates_id'],
				'name'                 => $description,
				'priority'             => $oscTaxRate['tax_priority'],
				'product_tax_classes'  => array(
					'id'  => $oscTaxRate['tax_class_id'],
					'key' => $oscTaxRate['tax_class_title']
				), // one class per rule in osCommerce
				'customer_tax_classes' => array(
					'id'  => 1,
					'key' => 'default'
				), // no customer tax classes in osCommerce
				'tax_rates'            => array(
					'id'  => $oscTaxRate['tax_rates_id'],
					'key' => $oscTaxRate['tax_rates_id']
				) // one rate per rule (since rates are in fact also rules) in osCommerce
			);
		}

		return array(
			'tax' => array(
				'product_tax_classes' => $this->getTaxClasses(), 'customer_tax_classes' => array(
					'id'         => 1,
					'key'        => 'default',
					'is_default' => 1
				), // no customer tax classes in osCommerce
				'tax_rates'           => $taxRates,
				'tax_rules'           => $taxRules,
			),
		);
	}

	/**
	 * @return mixed[]|void
	 */
	public function createShopInfo()
	{
		$shopInfo = array();

		$productCountQuery      = "SELECT count(*) cnt FROM " . TABLE_PRODUCTS . " as p WHERE p.products_status = 1";
		$result                 = ShopgateWrapper::db_query_print_on_err($productCountQuery);
		$row                    = ShopgateWrapper::db_fetch_array($result);
		$shopInfo['item_count'] = $row['cnt'];

		$catQry                     = "SELECT count(*) cnt FROM " . TABLE_CATEGORIES;
		$result                     = ShopgateWrapper::db_query_print_on_err($catQry);
		$row                        = ShopgateWrapper::db_fetch_array($result);
		$shopInfo['category_count'] = $row['cnt'];

		$reviewColumnExistQuery
				= "SELECT count(1) AS cnt FROM INFORMATION_SCHEMA.COLUMNS
									WHERE TABLE_SCHEMA = '" . DB_DATABASE
				  . "' AND TABLE_NAME = '" . TABLE_REVIEWS
				  . "' AND COLUMN_NAME = 'reviews_status'";
		$result = ShopgateWrapper::db_query_print_on_err($reviewColumnExistQuery);
		$row    = ShopgateWrapper::db_fetch_array($result);

		$revQry                   = "SELECT COUNT(*) AS cnt FROM " . TABLE_REVIEWS . " as r "
									. ($row["cnt"] > 0 ? " WHERE r.reviews_status = 1" : "");
		$result                   = ShopgateWrapper::db_query_print_on_err($revQry);
		$row                      = ShopgateWrapper::db_fetch_array($result);
		$shopInfo['review_count'] = $row['cnt'];

		// Not provided by Osc
		$shopInfo['plugins_installed '] = array();

		return $shopInfo;
	}

	/**
	 * @param ShopgateCart $cart
	 *
	 * @return array
	 */
	public function checkCart(ShopgateCart $cart)
	{
		global $customer_id;
		$customer_id                = $cart->getExternalCustomerId();
		$result["external_coupons"] = $this->checkCoupon($cart);
		$result["shipping_methods"] = $this->getShipping($cart);
		$result["currency"]         = $this->config->getCurrency();
		$result["items"]            = $this->checkValidityOfCartItems($cart);

		return $result;
	}

	/**
	 * increments the coupons quantity if coupon is valid
	 *
	 * @param ShopgateCart $cart
	 *
	 * @return ShopgateCoupon[]
	 */
	public function redeemCoupons(ShopgateCart $cart)
	{
		include_once(rtrim(dirname(__FILE__), "/") . "/model/ShopgateCouponModel.php");
		$coupons = $this->checkCoupon($cart);

		foreach ($coupons as &$coupon) {
			/**  @var ShopgateExternalCoupon $coupon */
			if ($coupon->getIsValid()) {
				if (!$this->couponModel->redeemValidCouponByCode($coupon->getCode())) {
					$coupon->setIsValid(false);
					$coupon->setNotValidMessage("Coupon quantity reached");
				}
			}
		}

		return $coupons;
	}

	/**
	 * @param string $jobname
	 * @param        $params
	 * @param string $message
	 * @param int    $errorcount
	 *
	 * @throws ShopgateLibraryException
	 */
	public function cron($jobname, $params, &$message, &$errorcount)
	{
		switch (strtoupper($jobname)) {
			case 'SET_SHIPPING_COMPLETED':
				$this->cronSetOrdersShippingCompleted($message, $errorcount);
				break;
			case 'CANCEL_ORDERS':
				$this->cronCancelOrder($message, $errorcount);
				break;
			default:
				throw new ShopgateLibraryException(
					ShopgateLibraryException::PLUGIN_CRON_UNSUPPORTED_JOB,
					'Job name: "' . $jobname . '"', true
				);
		}
	}

	############################################################################
	## UPDATE ORDER                                                           ##
	############################################################################

	/**
	 * Checks the items array and returns stock quantity for each item.
	 *
	 *
	 * @see http://wiki.shopgate.com/Shopgate_Plugin_API_check_cart#API_Response
	 *
	 * @param ShopgateCart $cart The ShopgateCart object to be checked and validated.
	 *
	 * @return array(
	 *          'items' => array(...), # list of item changes
	 * )
	 * @throws ShopgateLibraryException if an error occurs.
	 */
	public function checkStock(ShopgateCart $cart)
	{
		return array("items" => $this->checkValidityOfCartItems($cart));
	}

	/**
	 * Loads the Media file information to the products of the shop system's database and passes them to the buffer.
	 *
	 * Use ShopgatePlugin::buildDefaultMediaRow() to get the correct indices for the field names in a Shopgate media csv and
	 * use ShopgatePlugin::addMediaRow() to add it to the output buffer.
	 *
	 * @see http://wiki.shopgate.com/CSV_File_Media#Sample_Media_CSV_file
	 * @see http://wiki.shopgate.com/Shopgate_Plugin_API_get_media_csv
	 *
	 * @throws ShopgateLibraryException
	 */
	protected function createMediaCsv()
	{
		// TODO: Implement createMediaCsv() method.
	}

	/**
	 * Exports orders from the shop system's database to Shopgate.
	 *
	 * @see http://wiki.shopgate.com/Shopgate_Plugin_API_get_orders
	 *
	 * @param string $customerToken
	 * @param string $customerLanguage
	 * @param int    $limit
	 * @param int    $offset
	 * @param string $orderDateFrom
	 * @param string $sortOrder
	 *
	 * @return ShopgateExternalOrder[] A list of ShopgateExternalOrder objects
	 *
	 * @throws ShopgateLibraryException
	 */
	public function getOrders(
		$customerToken, $customerLanguage, $limit = 10, $offset = 0, $orderDateFrom = '', $sortOrder = 'created_desc'
	) {
		return $this->orderModel->getCustomerOrders($customerToken, $customerLanguage, $limit, $offset, $orderDateFrom, $sortOrder);
	}

	/**
	 * Updates and returns synchronization information for the favourite list of a customer.
	 *
	 * @see http://wiki.shopgate.com/Shopgate_Plugin_API_sync_favourite_list
	 *
	 * @param string             $customerToken
	 * @param ShopgateSyncItem[] $items A list of ShopgateSyncItem objects that need to be synchronized
	 *
	 * @return ShopgateSyncItem[] The updated list of ShopgateSyncItem objects
	 */
	public function syncFavouriteList($customerToken, $items)
	{
		// TODO: Implement syncFavouriteList() method.
	}

	/**
	 * @param $orderId
	 * @param int $ordersStatusId
	 * @param string $commentText
	 * @param string $forcedDate
	 * @return bool
	 * @throws ShopgateLibraryException
	 */
	private function saveSingleOrdersSatusHistory($orderId, $ordersStatusId = 1, $commentText = '', $forcedDate = '') {
		if (empty($forcedDate)) {
			$forcedDate = 'now()';
		}
		$ordersStatusHistory = array('orders_id' => $orderId, 'orders_status_id' => $ordersStatusId, 'date_added' => $forcedDate, 'customer_notified' => 0, 'comments' => $this->stringFromUtf8($commentText, $this->config->getEncoding()),);

		// Save entry to DB
		$queryResult = ShopgateWrapper::db_execute_query(TABLE_ORDERS_STATUS_HISTORY, $ordersStatusHistory);
		if (!$queryResult) {
			// DB-Error
			throw new ShopgateLibraryException(ShopgateLibraryException::PLUGIN_DATABASE_ERROR, "Shopgate Plugin - Error on inserting an orders status history entry.", true);
		}

		return true;
	}
	
	/**
	 * Checks if the given table exists
	 *
	 * @param string $tableName
	 *
	 * @throws ShopgateLibraryException
	 * @return boolean
	 */
	private function tableExists($tableName)
	{
		$tableName = trim($tableName);
		if (empty($tableName)) {
			return false;
		}

		// Get all table names
		$query = ShopgateWrapper::db_query("SHOW TABLES");
		if (!$query) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error checking for table \"$tableName\".",
				true
			);
		}
		while ($array = ShopgateWrapper::db_fetch_array($query)) {
			$array = array_values($array);

			// Check for table name
			if ($array[0] == $tableName) {
				return true;
			}
		}

		// The requested table has not been found if execution reaches here
		return false;
	}

	/**
	 * Takes a shopgate order from the library (using Merchant API [2.0]) and creates a new dataset for that order.
	 * After saving it to the db the full order dataset including the id is returned as an array.
	 * The parameters $statusHistoryComment and $isShippingBlockedComment will be filled with the information that is needed
	 * to be put to the order status history comment
	 *
	 * @param ShopgateOrder $oShopgateOrder (UTF-8 encoded data)
	 * @param               $currency
	 * @param               $ordersStatusId
	 * @param :out $statusHistoryComment
	 * @param :out $isShippingBlockedComment
	 *
	 * @return array
	 * @throws ShopgateLibraryException
	 */
	private function saveNewOrder(
		ShopgateOrder $oShopgateOrder, $currency, &$ordersStatusId, &$statusHistoryComment, &$isShippingBlockedComment
	) {
		// Get the customer and his address (create a dataset if there is no customer found)
		$customerData = $this->getCustomerData($oShopgateOrder);
		// getCustomerData returns a 2-dimensional array: array('customer' => $dbCustomerData, 'customers_address' => $dbCustomersAddress)

		// Get a valid customers email address to save for the order (guest users get a "guest user" email address, so here the real email address has to be taken)
		if (empty($customerData['customer']['customers_email_address'])
			|| $customerData['customer']['customers_email_address'] == $this->createGuestUserEmail(
				$oShopgateOrder->getOrderNumber()
			)
		) {
			$customersEmailAddress = $oShopgateOrder->getMail(); // Take the email address given by shopgate
		} else {
			$customersEmailAddress = $customerData['customer']['customers_email_address'];
		}

		// -> Get the addresses (in UTF-8 format) and convert to the intern format
		/** @var ShopgateAddress $billingAddress */
		/** @var ShopgateAddress $deliveryAddress */
		$billingAddress    = $oShopgateOrder->getInvoiceAddress()->utf8Decode($this->config->getEncoding());
		$deliveryAddress   = $oShopgateOrder->getDeliveryAddress()->utf8Decode($this->config->getEncoding());
		$billingStateCode  = $billingAddress->getState();
		$deliveryStateCode = $deliveryAddress->getState();

		// -> Get delivery and billing zone and country (do only multiple selects if the two addresses differ)
		$billingCountry = $this->getCountry(
			$billingAddress->getCountry(),
			array('countries_id', 'countries_name', 'address_format_id')
		);
		if (empty($billingCountry)) {
			// Missing database-entry
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_UNKNOWN_COUNTRY_CODE,
				"Shopgate Plugin - No country found with the iso-code-2 \""
				. $billingAddress->getCountry() . "\".", true
			);
		}
		$billingStateZone = array();
		if (!empty($billingStateCode)) {
			$billingStateZone = $this->getStateZone(
				$billingCountry['countries_id'], $billingStateCode,
				array('zone_id', 'zone_name')
			);
		}
		if ($billingAddress->getCountry() == $deliveryAddress->getCountry()) {
			// Countries are identical, so just take the same country
			$deliveryCountry = $billingCountry;

			// The state can still be different
			if ($billingStateCode == $deliveryStateCode) {
				$deliveryStateZone = $billingStateZone;
			} else {
				$deliveryStateZone = array();
				if (!empty($deliveryStateCode)) {
					$deliveryStateZone = $this->getStateZone(
						$deliveryCountry['countries_id'], $deliveryStateCode,
						array('zone_id', 'zone_name')
					);
				}
			}
		} else {
			// Countries differ
			$deliveryCountry = $this->getCountry(
				$deliveryAddress->getCountry(),
				array('countries_id', 'countries_name', 'address_format_id')
			);
			if (empty($deliveryCountry)) {
				// Missing database-entry
				throw new ShopgateLibraryException(
					ShopgateLibraryException::PLUGIN_UNKNOWN_COUNTRY_CODE,
					"Shopgate Plugin - No country found with the iso-code-2 \""
					. $deliveryCountry->getCountry() . "\".", true
				);
			}

			// Different country means ALWAYS a different state!
			$deliveryStateZone = array();
			if (!empty($deliveryStateCode)) {
				$deliveryStateZone = $this->getStateZone(
					$deliveryCountry['countries_id'], $deliveryStateCode,
					array('zone_id', 'zone_name')
				);
			}
		}

		// -> Get the customers country data
		if (!empty($customerData['customers_address'])) {
			$customersCountry = $this->getCountry(
				null /* get by countries_id, instead of isoCode2 */,
				array('countries_id', 'countries_name', 'address_format_id'),
				$customerData['customers_address']['entry_country_id']
			);
		}
		// -> Fallback for the case if no country is found
		if (empty($customersCountry)) {
			// Default the country to "German" or "United States" if none set (depending on plugin version)
			if (strpos(SHOPGATE_PLUGIN_VERSION, 'usa') === false) {
				$customersCountry
					= array(
					'countries_id'      => '81', 'countries_name' => 'Germany',
					'address_format_id' => '5'
				);
			} else {
				$customersCountry = array(
					'countries_id'      => '223',
					'countries_name'    => 'United States',
					'address_format_id' => '2'
				);
			}
		}

		// -> set orders status
		//	=>	1 = Pending;
		//	=>	2 = Processing;
		//	=>	3 = Delivered;
		//		There are two possible status that can be valid, depending on if the shipping is blocked or not
		if ($oShopgateOrder->getIsShippingBlocked()) {
			$ordersStatusId = $this->config->getOrderStatusShippingBlocked();
		} else {
			$ordersStatusId = $this->config->getOrderStatusOpen();
		}

		// On default "pending" status there is more information needed, that is set es comment to the status history
		$isShippingBlockedComment = '';
		if ($oShopgateOrder->getIsShippingBlocked()) {
			$isShippingBlockedComment
				= ENTRY_SHIPPING_BLOCKED_COMMENT_TEXT . "\n";
		} else {
			$isShippingBlockedComment
				= ENTRY_SHIPPING_NOT_BLOCKED_COMMENT_TEXT . "\n";
		}

		// Data is encoded using latin-1 here
		// -> Orders customer name
		$order['customers_id']                = $customerData['customer']['customers_id'];
		$order['customers_name']              = $customerData['customers_address']['entry_firstname'] . ' '
												. $customerData['customers_address']['entry_lastname'];
		$order['customers_company']           = $customerData['customers_address']['entry_company'];
		$order['customers_street_address']    = $customerData['customers_address']['entry_street_address'];
		$order['customers_suburb']            = $customerData['customers_address']['entry_suburb'];
		$order['customers_city']              = $customerData['customers_address']['entry_city'];
		$order['customers_postcode']          = $customerData['customers_address']['entry_postcode'];
		$order['customers_state']             = $customerData['customers_address']['entry_state'];
		$order['customers_country']           = $customersCountry['countries_name'];
		$order['customers_telephone']         = $customerData['customer']['customers_telephone'];
		$order['customers_email_address']     = $customersEmailAddress;
		$order['customers_address_format_id'] = $customersCountry['address_format_id'];
		// -> Orders delivery address
		$order['delivery_name']    = $deliveryAddress->getFirstName() . ' ' . $deliveryAddress->getLastName();
		$order['delivery_company'] = $deliveryAddress->getCompany();
		$order['delivery_street_address']
											 =
			$deliveryAddress->getStreet1() . (strlen($deliveryAddress->getStreet2()) > 0 ? (' '
																							. $deliveryAddress->getStreet2(
				)) : '');
		$order['delivery_suburb']            = ''; // Unsupported
		$order['delivery_city']              = $deliveryAddress->getCity();
		$order['delivery_postcode']          = $deliveryAddress->getZipcode();
		$order['delivery_state']             = (!empty($deliveryStateZone) ? $deliveryStateZone['zone_name'] : "");
		$order['delivery_country']           = $deliveryCountry['countries_name'];
		$order['delivery_address_format_id'] = $deliveryCountry['address_format_id'];
		// -> Orders billing address
		$order['billing_name']    = $billingAddress->getFirstName() . ' ' . $deliveryAddress->getLastName();
		$order['billing_company'] = $billingAddress->getCompany();
		$order['billing_street_address']
											=
			$billingAddress->getStreet1() . (strlen($billingAddress->getStreet2()) > 0 ? (' '
																						  . $billingAddress->getStreet2(
				)) : '');
		$order['billing_suburb']            = ''; // Unsupported
		$order['billing_city']              = $billingAddress->getCity();
		$order['billing_postcode']          = $billingAddress->getZipcode();
		$order['billing_state']             = (!empty($billingStateZone) ? $billingStateZone['zone_name'] : "");
		$order['billing_country']           = $billingCountry['countries_name'];
		$order['billing_address_format_id'] = $billingCountry['address_format_id'];
		// -> Payment information
		$statusHistoryComment = "";
		$paymentInfo          = $oShopgateOrder->getPaymentInfos();
		switch ($oShopgateOrder->getPaymentMethod()) {
			case ShopgateOrder::SHOPGATE:
				$order['payment_method'] = ENTRY_PAYMENT_SHOPGATE;
				break;
			case ShopgateOrder::PREPAY:
				$order['payment_method'] = ENTRY_PAYMENT_PREPAY;
				// Save additional information into order status history comment
				$statusHistoryComment .= ENTRY_PAYMENT_PREPAY_PAYMENT_PURPOSE . ' ' . $paymentInfo['purpose'] . "\n";
				break;
			case ShopgateOrder::DEBIT:
				// -> not directly supported by osCommerce; use a generic import method
				$order['payment_method'] = ENTRY_PAYMENT_SHOPGATE_GENERIC;
				// Save additional information into status history comments
				$statusHistoryComment .= ENTRY_PAYMENT_ELV . ":\n\n";
				$statusHistoryComment .= ENTRY_PAYMENT_ELV_INFOTEXT . "\n\n";
				$statusHistoryComment
					.= ENTRY_PAYMENT_ELV_TEXT_BANK_ACCOUNT_OWNER . " " . $paymentInfo["bank_account_holder"] . "\n";
				$statusHistoryComment
					.= ENTRY_PAYMENT_ELV_TEXT_BANK_ACCOUNT_NUMBER . " " . $paymentInfo["bank_account_number"] . "\n";
				$statusHistoryComment .= ENTRY_PAYMENT_ELV_TEXT_BANK_CODE . " " . $paymentInfo["bank_code"] . "\n";
				$statusHistoryComment .= ENTRY_PAYMENT_ELV_TEXT_BANK_NAME . " " . $paymentInfo["bank_name"] . "\n";
				break;
			case ShopgateOrder::COD:
				$order['payment_method'] = ENTRY_PAYMENT_CASH_ON_DELIVERY;
				break;
			default:
				// Use a generic payment type for all other payment methods, that are not directly supported by shoppingsystem
				$order['payment_method'] = ENTRY_PAYMENT_SHOPGATE_GENERIC;
				// Put all available information as comment into the order status
				foreach ($paymentInfo as $key => $value) {
					$statusHistoryComment .= $key . ': ' . $value . "\n";
				}
				break;
		}
		// -> Credit card settings (unsupported)
		$order['cc_owner']   = "";
		$order['cc_type']    = "";
		$order['cc_number']  = "";
		$order['cc_expires'] = "";
		// -> Various extra information
		$order['last_modified']  = date('Y-m-d H:i:s');
		$order['date_purchased'] = date('Y-m-d H:i:s', strtotime($oShopgateOrder->getCreatedTime()));
		$order['orders_status']  = $ordersStatusId;
		$order['currency']       = !empty($currency['code']) ? $currency['code'] : '';
		$order['currency_value'] = !empty($currency['value']) ? $currency['value'] : 1;

		// -> Insert the order to the database
		$queryResult = ShopgateWrapper::db_execute_query(TABLE_ORDERS, $order);
		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error on inserting an order dataset.", true
			);
		}
		$order['orders_id'] = ShopgateWrapper::db_insert_id();

		$order['billing_firstname'] = $billingAddress->getFirstName();
		$order['billing_lastname'] = $billingAddress->getLastName();
		$order['billing_country_id'] = $billingCountry['countries_id'];
		$order['billing_zone_id'] = $billingCountry['zone_id'];

		$order['delivery_firstname'] = $deliveryAddress->getFirstName();
		$order['delivery_lastname'] = $deliveryAddress->getLastName();
		$order['delivery_country_id'] = $deliveryCountry['countries_id'];
		$order['delivery_zone_id'] = $deliveryStateZone['zone_id'];

		return $order;
	}

	/**
	 * Takes the order data and creates a customer dataset.
	 * After finishing it returns all customer data for a new order to be created
	 *
	 * @param ShopgateOrder $oShopgateOrder (UTF-8 encoded data)
	 *
	 * @return array[][]
	 * @throws ShopgateLibraryException
	 */
	private function getCustomerData(ShopgateOrder $oShopgateOrder)
	{
		$dbCustomerData     = array();
		$externalCustomerId = $oShopgateOrder->getExternalCustomerId();
		if (!empty($externalCustomerId)) {
			// Check if there is a customer with the given id set
			$sqlQuery = "SELECT\n" . "\t*\n" . "FROM\n" . "\t" . TABLE_CUSTOMERS
						. " AS customer\n" . "WHERE\n"
						. "\tcustomer.customers_id={$externalCustomerId}\n" . ";";

			// -> Get results
			$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);
			if (!$queryResult) {
				throw new ShopgateLibraryException(
					ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
					"Shopgate Plugin - Error selecting shopgate orders.", true
				);
			}
			$dbCustomerData = ShopgateWrapper::db_fetch_array($queryResult);
		}

		// Create a new guest-account if there is none found
		if (empty($dbCustomerData)) {
			/** @var ShopgateAddress $customerAddress */
			// Use the invoice address, since it is only possible to pass one address as user data
			$customerAddress = $oShopgateOrder->getInvoiceAddress();
			$customerAddress = $customerAddress->utf8Decode(
				$this->config->getEncoding()
			);

			// Fill the customers data
			$newCustomerData                                 = array();
			$newCustomerData['customers_gender']             = $customerAddress->getGender();
			$newCustomerData['customers_firstname']          = $customerAddress->getFirstName();
			$newCustomerData['customers_lastname']           = $customerAddress->getLastName();
			$newCustomerData['customers_dob']                = $customerAddress->getBirthday();
			$newCustomerData['customers_email_address']      = $this->createGuestUserEmail(
				$oShopgateOrder->getOrderNumber()
			);
			$newCustomerData['customers_default_address_id'] = '0'; // Unknown yet
			$newCustomerData['customers_telephone']          = $oShopgateOrder->getPhone();
			$newCustomerData['customers_fax']                = ''; // Unsupported
			$newCustomerData['customers_password']           = ''; // Stays empty. Guest customers can't log in
			$newCustomerData['customers_newsletter']         = '0'; // Do not activate newsletter

			// Create the guest-customer
			$queryResult = ShopgateWrapper::db_execute_query(TABLE_CUSTOMERS, $newCustomerData);
			if (!$queryResult) {
				throw new ShopgateLibraryException(
					ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
					"Shopgate Plugin - Error on inserting a customer dataset.",
					true
				);
			}

			// Set the id
			$newCustomerData['customers_id'] = ShopgateWrapper::db_insert_id();

			// Create a customers_info dataset
			$newCustomerInfo = array(
				'customers_info_id'                   => $newCustomerData['customers_id'],
				'customers_info_number_of_logons'     => 0,
				'customers_info_date_account_created' => date('Y-m-d H:i:s'),
				'global_product_notifications'        => '0'
			);
			$queryResult     = ShopgateWrapper::db_execute_query(
				TABLE_CUSTOMERS_INFO, $newCustomerInfo
			);
			if (!$queryResult) {
				throw new ShopgateLibraryException(
					ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
					"Shopgate Plugin - Error on inserting a customers info-dataset.",
					true
				);
			}

			// Create a new address book entry
			$newCustomersAddress                    = array();
			$newCustomersAddress['customers_id']    = $newCustomerData['customers_id'];
			$newCustomersAddress['entry_gender']    = $customerAddress->getGender();
			$newCustomersAddress['entry_company']   = $customerAddress->getCompany();
			$newCustomersAddress['entry_firstname'] = $customerAddress->getFirstName();
			$newCustomersAddress['entry_lastname']  = $customerAddress->getLastName();
			$newCustomersAddress['entry_street_address']
													=
				$customerAddress->getStreet1() . (strlen($customerAddress->getStreet2()) > 0 ? (' '
																								. $customerAddress->getStreet2(
					)) : '');
			$newCustomersAddress['entry_suburb']    = '';
			$newCustomersAddress['entry_postcode']  = $customerAddress->getZipcode();
			$newCustomersAddress['entry_city']      = $customerAddress->getCity();
			// Add additional data (country_id, zone_id and state)
			$countryData = $this->getCountry($customerAddress->getCountry(), array('countries_id'));
			if (!empty($countryData)) {
				$stateData = $this->getStateZone(
					$countryData['countries_id'], $customerAddress->getState(), array('zone_id', 'zone_name')
				);

				$newCustomersAddress['entry_country_id'] = $countryData['countries_id'];
				if (!empty($stateData)) {
					$newCustomersAddress['entry_state']   = $stateData['zone_name'];
					$newCustomersAddress['entry_zone_id'] = $stateData['zone_id'];
				}
			}

			$queryResult = ShopgateWrapper::db_execute_query(TABLE_ADDRESS_BOOK, $newCustomersAddress);
			if (!$queryResult) {
				throw new ShopgateLibraryException(
					ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
					"Shopgate Plugin - Error on inserting a customer address.",
					true
				);
			}
			$newCustomersAddress['address_book_id'] = ShopgateWrapper::db_insert_id();

			// Set default address for new customer
			$updateCustomer = array('customers_default_address_id' => $newCustomersAddress['address_book_id']);
			$queryResult    = ShopgateWrapper::db_execute_query(
				TABLE_CUSTOMERS, $updateCustomer, 'update',
				"customers_id = $newCustomersAddress[customers_id]"
			);
			if (!$queryResult) {
				throw new ShopgateLibraryException(
					ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
					"Shopgate Plugin - Error on updating a customers default address.",
					true
				);
			}

			// Fill the return array
			$customerData = array(
				'customer'          => $newCustomerData,
				'customers_address' => $newCustomersAddress
			);
		} else {
			// Get customers address
			$sqlQuery    = "SELECT\n" . "\t*\n" . "FROM `" . TABLE_ADDRESS_BOOK
						   . "` AS `ab`\n"
						   . "WHERE `ab`.`customers_id` = '{$dbCustomerData['customers_id']}'\n"
						   . (!empty($dbCustomerData['customers_default_address_id'])
					? ("\tAND `ab`.`address_book_id` = '{$dbCustomerData['customers_default_address_id']}'\n")
					: ("")) . ";";
			$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);
			if (!$queryResult) {
				throw new ShopgateLibraryException(
					ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
					"Shopgate Plugin - Error on selecting the customers default address.",
					true
				);
			}
			$dbCustomersAddress = ShopgateWrapper::db_fetch_array($queryResult);

			// Fill the return array
			$customerData = array(
				'customer'          => $dbCustomerData,
				'customers_address' => $dbCustomersAddress
			);
		}

		return $customerData;
	}

	/**
	 * Creates a shopgate guest user email address based on the shopgate order number
	 *
	 * @param string $shopgateOrderNumber
	 *
	 * @return string
	 */
	private function createGuestUserEmail($shopgateOrderNumber)
	{
		return "noreply+{$shopgateOrderNumber}@shopgate.com";
	}

	/**
	 * Takes a country code in "ISO-3166-1 ALPHA-2" - format and an array of all needed fields and returns the
	 * corresponding data that is saved in the database
	 *
	 * @param string $isoCode2
	 * @param array  $fields
	 * @param int    $id (optional)
	 *
	 * @throws ShopgateLibraryException
	 * @return integer
	 */
	private function getCountry($isoCode2, $fields = array(), $id = null)
	{
		// Get the desired fields id from databases using the given iso code
		if (empty($fields)) {
			$elements = 'country.*';
		} else {
			$elements = 'country.' . implode(",\ncountry.", $fields);
		}
		$sqlQuery
			= "SELECT\n" . "\t$elements\n" . "FROM\n" . "\t" . TABLE_COUNTRIES
			  . " AS country\n" . "WHERE\n" . (!empty($id)
				? "country.countries_id=$id"
				: "\tUPPER(country.countries_iso_code_2)='" . addslashes(
					$isoCode2
				) . "'\n") . ";";

		// -> Get results
		$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);
		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error selecting country.", true
			);
		}

		return ShopgateWrapper::db_fetch_array($queryResult);
	}

	/**
	 * Takes a country_id and a zone code in "ISO 3166-2" - format and an array with all fields needed and returns the
	 * corresponding zone_id that is saved in the database
	 *
	 * @param integer $countryId
	 * @param string  $shopgateZoneCode
	 * @param array   $fields
	 *
	 * @throws ShopgateLibraryException
	 * @return integer
	 */
	private function getStateZone($countryId, $shopgateZoneCode, $fields = array())
	{
		if (!empty($countryId) && !empty($shopgateZoneCode)) {
			// Convert the state code to system format
			$oscStateCode = ShopgateMapper::getShoppingsystemStateCode($shopgateZoneCode);

			// Load the desired fields id from the database, using the corresponding state-code and the previously selected country id
			if (empty($fields)) {
				$elements = 'zone.*';
			} else {
				$elements = 'zone.' . implode(",\nzone.", $fields);
			}
			$sqlQuery = "SELECT\n" . $elements . ' ' . "FROM\n" . "\t" . TABLE_ZONES
						. " AS zone\n" . "WHERE\n" . "\tzone.zone_code='" . addslashes($oscStateCode)
						. "'\n" . "\t\tAND\n" . "\tzone.zone_country_id=" . intval($countryId) . "\n" . ";";

			// -> Get results
			$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);
			if (!$queryResult) {
				throw new ShopgateLibraryException(
					ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
					"Shopgate Plugin - Error selecting state zone.", true
				);
			}

			$zoneData = ShopgateWrapper::db_fetch_array($queryResult);

			return $zoneData;
		}

		// By default there is no valid zone returned
		return array();
	}

	/**
	 * Creates a shopgate-order dataset and saves it to the database. After saving the dataset is returned including the inserted id.
	 *
	 * @param ShopgateOrder $oShopgateOrder (UTF-8 encoded data)
	 * @param array         $orderId
	 *
	 * @throws ShopgateLibraryException
	 * @return array
	 */
	private function saveShopgateOrder(ShopgateOrder $oShopgateOrder, $orderId)
	{
		$shopgateOrder = array(
			'orders_id'                        => $orderId,
			'shopgate_order_number'            => $oShopgateOrder->getOrderNumber(),
			'shopgate_shop_number'             => $this->config->getShopNumber(),
			'is_paid'                          => $oShopgateOrder->getIsPaid(),
			'is_shipping_blocked'              => $oShopgateOrder->getIsShippingBlocked(),
			'order_data'                       => serialize($oShopgateOrder),
			'payment_infos'                    => $this->jsonEncode($oShopgateOrder->getPaymentInfos(), true),
			'is_sent_to_shopgate'              => 0, // No "shipped" request is sent on adding an order
			'is_cancellation_sent_to_shopgate' => 0,
			'created'                          => date('Y-m-d H:i:s')
		);

		// -> Insert the order to the database
		$queryResult = ShopgateWrapper::db_execute_query(TABLE_ORDERS_SHOPGATE_ORDER, $shopgateOrder);
		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error on inserting a shopgate-order dataset.",
				true
			);
		}
		$shopgateOrder['shopgate_order_id'] = ShopgateWrapper::db_insert_id();

		return $shopgateOrder;
	}

	/**
	 * Adds all ordered items into the orders_products table
	 *
	 * @param ShopgateOrder $oShopgateOrder (UTF-8 encoded data)
	 * @param               $orderId
	 * @param               $currency
	 * @param               $errors
	 *
	 * @return array
	 * @throws ShopgateLibraryException
	 */
	private function saveOrdersItems(ShopgateOrder $oShopgateOrder, $orderId, $currency, &$errors)
	{
		$errors       = array();
		$ordersItems  = array();
		$orderedItems = $oShopgateOrder->getItems();
		$i            = 0;
		foreach ($orderedItems as $singleItem) {
			// check if the product exists
			$sqlQuery
						 = "SELECT COUNT(*) AS `product_count`, `products_price` FROM `"
						   . TABLE_PRODUCTS . "` WHERE `" . TABLE_PRODUCTS
						   . "`.`products_id` = '" . intval($singleItem->getItemNumber())
						   . "' GROUP BY `products_id`";
			$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);
			if (!$queryResult) {
				throw new ShopgateLibraryException(
					ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
					"Shopgate Plugin - Error on selecting a product.", true
				);
			}
			$dbProduct = ShopgateWrapper::db_fetch_array($queryResult);
			if (empty($dbProduct) || $dbProduct['product_count'] <= 0) {
				$dbProduct['product_count'] = 0;
			}
			$productsModel = $singleItem->getItemNumberPublic();

			// Calculate the products price under consideration of all options
			$optionsUpPrice = 0;
			$itemTaxPercent = $singleItem->getTaxPercent() + 100;
			$itemOptions    = $singleItem->getOptions();
			if (!empty($itemOptions)) {
				foreach ($itemOptions as $itemOption) {
					/* @var $itemOption ShopgateOrderItemOption */
					// Remove taxes here! -> the amount is a shopgate-internal-amount (integer type)
					$optionsUpPrice += $itemOption->getAdditionalAmountWithTax() / $itemTaxPercent * 100;
				}
			}

			// Remove the options prices from the complete price to get the product price
			$productsPrice    = $singleItem->getUnitAmountWithTax() / $itemTaxPercent * 100 - $optionsUpPrice;
			$currencyExchange = !empty($currency['value']) ? round($currency['value'], 6) : 1;

			// Build the ordered items structure
			$ordersItems[$i]['orders_id']      = $orderId;
			$ordersItems[$i]['products_id']    = intval($singleItem->getItemNumber());
			$ordersItems[$i]['products_model'] = $this->stringFromUtf8($productsModel, $this->config->getEncoding());
			$ordersItems[$i]['products_name']  = $this->stringFromUtf8(
				$singleItem->getName(), $this->config->getEncoding()
			);
			$ordersItems[$i]['products_price']
												  =
				$productsPrice / $currencyExchange; // products-price without tax and without options
			$ordersItems[$i]['final_price']       = $singleItem->getUnitAmountWithTax() / $itemTaxPercent * 100
													/ $currencyExchange; // Unit amount already contains the full amount including the options options price offset
			$ordersItems[$i]['products_tax']      = $singleItem->getTaxPercent();
			$ordersItems[$i]['products_quantity'] = $singleItem->getQuantity();

			// Check if price from db product is nearly the same to get a more exact value
			if (!empty($dbProduct['products_price'])
				&& round($dbProduct['products_price'], 1) == round($ordersItems[$i]['products_price'], 1)
			) {
				// Take price from database in this case
				$ordersItems[$i]['products_price'] = $dbProduct['products_price'];
				$ordersItems[$i]['final_price']    = $dbProduct['products_price'];

				// Check if the actual product has options
				if (!empty($itemOptions)) {
					$additionalAmount = 0;
					foreach ($singleItem->getOptions() as $itemOption) {
						$additionalAmount += round(
							$itemOption->getAdditionalAmountWithTax() / $itemTaxPercent * 100 / $currencyExchange, 2
						);
					}
					$ordersItems[$i]['final_price'] += $additionalAmount;
				}
			}

			// Coupons and payment fees are exported as single item with a special item number
			if ($singleItem->getItemNumber() == 'COUPON'
				|| $singleItem->getItemNumber() == 'PAYMENT_FEE'
				|| strlen(preg_replace('/^[+-]?[0-9]*$/', '', $singleItem->getItemNumber())) > 0
			) {
				$ordersItems[$i]['products_id']    = 0;
				$ordersItems[$i]['products_model'] = $this->stringFromUtf8(
					$singleItem->getItemNumber(), $this->config->getEncoding()
				);
			} elseif ($dbProduct['product_count'] == 0) { // Product not found!
				$ordersItems[$i]['products_id']    = 0;
				$ordersItems[$i]['products_model'] = $this->stringFromUtf8(
					$singleItem->getItemNumber(), $this->config->getEncoding()
				);

				// Save an error message
				$errors[] = str_replace(
					'#PRODUCTS_ID#', $this->stringFromUtf8(
						$singleItem->getItemNumber(),
						$this->config->getEncoding()
					), ENTRY_PRODUCT_NOT_FOUND
				);
			} else {
				// Update products quantity
				$updateSql   = "UPDATE `" . TABLE_PRODUCTS
							   . "` SET `products_quantity` = `products_quantity`-{$singleItem->getQuantity()} WHERE `products_id` = '"
							   . intval($singleItem->getItemNumber(), 10) . "'";
				$queryResult = ShopgateWrapper::db_query_print_on_err($updateSql);
				if (!$queryResult) {
					throw new ShopgateLibraryException(
						ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
						"Shopgate Plugin - Error on updating a products quantity.",
						true
					);
				}
			}

			// -> Insert the order to the database
			$queryResult = ShopgateWrapper::db_execute_query(TABLE_ORDERS_PRODUCTS, $ordersItems[$i]);
			if (!$queryResult) {
				throw new ShopgateLibraryException(
					ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
					"Shopgate Plugin - Error on inserting an orders product.",
					true
				);
			}
			$ordersItems[$i]['orders_products_id']
				= ShopgateWrapper::db_insert_id();

			// Insert all selected options
			$itemOptionsList = array();
			if (!empty($itemOptions)) {
				foreach ($singleItem->getOptions() as $itemOption) {
					/* @var $itemOption ShopgateOrderItemOption */
					$optionPricePrefix = '+';
					if ($itemOption->getAdditionalAmountWithTax() < 0) {
						$optionPricePrefix = '-';
					}

					// create options structure array
					$newItemOption = array(
						'orders_id'               => $orderId,
						'orders_products_id'      => $ordersItems[$i]['orders_products_id'],
						'products_options'        => $itemOption->getName(),
						'products_options_values' => $itemOption->getValue(),
						'options_values_price'    => round(
							abs($itemOption->getAdditionalAmountWithTax()) / $itemTaxPercent * 100 / $currencyExchange,
							2
						), // no price-prefix here!
						'price_prefix'            => $optionPricePrefix
					);

					// Add to database
					$queryResult = ShopgateWrapper::db_execute_query(TABLE_ORDERS_PRODUCTS_ATTRIBUTES, $newItemOption);
					if (!$queryResult) {
						throw new ShopgateLibraryException(
							ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
							"Shopgate Plugin - Error on inserting an orders product attribute.",
							true
						);
					}
					$newItemOption['orders_products_attributes_id'] = ShopgateWrapper::db_insert_id();

					// Add option to list
					$itemOptionsList += $newItemOption;
				}
			}

			$i++;
		}

		return $ordersItems;
	}

	/**
	 * Calculates all amounts to a total price and adds all needed entries to the orders_total table
	 *
	 * @param ShopgateOrder $oShopgateOrder (UTF-8 encoded data)
	 * @param integer       $orderId
	 * @param array         $currency
	 * @param Array         $aOrdersItems
	 *
	 * @throws ShopgateLibraryException
	 * @return array
	 */
	private function saveOrdersTotal(ShopgateOrder $oShopgateOrder, $orderId, $currency, $aOrdersItems)
	{
		$oCurrencies          = new currencies();
		$currencyCode         = !empty($currency['code']) ? $currency['code'] : '';
		$currencyExchangeRate = !empty($currency['value']) ? $currency['value'] : 1;
		$valueShipping        = $oShopgateOrder->getAmountShipping();
		$valuePaymentCosts    = $oShopgateOrder->getAmountShopPayment();
		$shippingInfos        = $oShopgateOrder->getShippingInfos();
		$valueTaxes           = array();
		$valueTaxTotal        = 0;

		// Compute all values
		foreach ($aOrdersItems as $aSingleItem) {

			$positionValue      = $aSingleItem['final_price'] * $aSingleItem['products_quantity'];
			$positionTaxPercent = $aSingleItem['products_tax'];
			$positionTax        = $positionValue * $positionTaxPercent / 100;

			// Add up tax price value only
			if (empty($valueTaxes[$positionTaxPercent])) {
				$valueTaxes[$positionTaxPercent] = 0;
			}
			$valueTaxes[$positionTaxPercent] += $positionTax * $currencyExchangeRate;
		}

		foreach ($valueTaxes as $valueTax) {
			$valueTaxTotal += $valueTax;
		}

		$valueTotal = $oShopgateOrder->getAmountComplete();

		// Use better method for calculating sub total (without tax, payment costs and shipping)
		$valueSubTotal = $valueTotal - $valueTaxTotal - $valuePaymentCosts - $valueShipping;

		$ordersTotals = array();
		$i            = 0;
		//check coupons
		if (defined("MODULE_ORDER_TOTAL_INSTALLED")
			&& strpos(MODULE_ORDER_TOTAL_INSTALLED, "ot_discount_coupon.php") !== false
		) {
			$coupons = $oShopgateOrder->getExternalCoupons();

			foreach ($coupons as $coupon) {
				$valueSubTotal -= ($coupon->getAmount());
				$amount           = ($coupon->getAmount() * (-1));
				$ordersTotals[$i] = array(
					'orders_id'  => $orderId, 'title' => sprintf(ENTRY_COUPON, $coupon->getCode()),
					'text'       => $oCurrencies->format($amount, false, $currencyCode),
					'value'      => $amount,
					'class'      => 'ot_discount_coupon',
					'sort_order' => $i + 1,
				);
				$i++;
			}
		}

		// Add price without taxes and shipping costs
		$ordersTotals[++$i] = array(
			'orders_id'  => $orderId, 'title' => ENTRY_SUB_TOTAL,
			'text'       => $oCurrencies->format($valueSubTotal, false, $currencyCode),
			'value'      => $valueSubTotal, 'class' => 'ot_subtotal',
			'sort_order' => $i + 1,
		);

		$ordersTotals[++$i] = array(
			'orders_id'  => $orderId,
			'title'      => $shippingInfos->getDisplayName(),
			'text'       => $oCurrencies->format($valueShipping, false, $currencyCode),
			'value'      => $valueShipping,
			'class'      => 'ot_shipping',
			'sort_order' => $i + 1,
		);

		// Add payment costs
		if ($valuePaymentCosts > 0) {
			$ordersTotals[++$i]
				= array(
				'orders_id' => $orderId, 'title' => ENTRY_PAYMENT,
				'text'      => $oCurrencies->format($valuePaymentCosts, false, $currencyCode),
				'value'     => $valuePaymentCosts,
				'class'     => 'ot_payment', 'sort_order' => $i + 1,
			);
		}

		// Add taxes (separated per percent)
		foreach ($valueTaxes as $valueTaxPercent => $valueTax) {
			$ordersTotals[++$i] = array(
				'orders_id'  => $orderId, 'title' => sprintf(ENTRY_TAX, $valueTaxPercent),
				'text'       => $oCurrencies->format($valueTax, false, $currencyCode),
				'value'      => $valueTax, 'class' => 'ot_tax',
				'sort_order' => $i + 1,
			);
		}

		// Finally add the total value
		$ordersTotals[++$i]
			= array(
			'orders_id'  => $orderId, 'title' => ENTRY_TOTAL,
			'text'       => '<strong>' . ($oCurrencies->format($valueTotal, false, $currencyCode)) . '</strong>',
			'value'      => $valueTotal, 'class' => 'ot_total',
			'sort_order' => $i + 1,
		);

		// Add all lines to the database
		foreach ($ordersTotals as $key => $dbEntry) {
			$queryResult = ShopgateWrapper::db_execute_query(TABLE_ORDERS_TOTAL, $dbEntry);
			if (!$queryResult) {
				throw new ShopgateLibraryException(
					ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
					"Shopgate Plugin - Error on inserting an orders total entry.",
					true
				);
			}
			$ordersTotals[$key]['orders_total_id'] = ShopgateWrapper::db_insert_id();
		}

		return $ordersTotals;
	}

	/**
	 * Builds status history entries out of given comment data and inserts them into the database
	 *
	 * @param ShopgateOrder $oShopgateOrder
	 * @param               $orderId
	 * @param int           $isGenericPaymentInfo
	 * @param int           $ordersStatusId
	 * @param string        $statusHistoryComment
	 * @param string        $isShippingBlockedComment
	 * @param array         $errors
	 * @param string        $forcedDate
	 *
	 * @return bool
	 * @throws ShopgateLibraryException
	 */
	private function saveOrdersStatusHistory(
		ShopgateOrder $oShopgateOrder, $orderId, $isGenericPaymentInfo = 0,
		$ordersStatusId = 1, $statusHistoryComment = '',
		$isShippingBlockedComment = '', $errors = array(), $forcedDate = ''
	) {
		if (empty($forcedDate)) {
			$forcedDate = 'now()';
		}
		$ordersStatusHistory = array('orders_id'         => $orderId,
									 'orders_status_id'  => $ordersStatusId,
									 'date_added'        => $forcedDate,
									 'customer_notified' => 0,
									 'comments'          => '');

		// Setup comment
		if ($oShopgateOrder->getIsTest()) {
			$ordersStatusHistory['comments'] .= $this->stringFromUtf8(
				ENTRY_IS_TEST_ORDER_COMMENT_TEXT . "\n",
				$this->config->getEncoding()
			);
		}
		// -> added by shopgate comment text
		if (!empty($ordersStatusHistory['comments'])) {
			$ordersStatusHistory['comments'] .= "\n";
		}
		$ordersStatusHistory['comments'] .= $this->stringFromUtf8(
			ENTRY_ORDER_ADDED_BY_SHOPGATE_COMMENT_TEXT . "\n",
			$this->config->getEncoding()
		);
		// Insert shopgate order-number
		$shopgateOrderNumber = $oShopgateOrder->getOrderNumber();
		if (!empty($shopgateOrderNumber)) {
			$ordersStatusHistory['comments'] .= $this->stringFromUtf8(
				ENTRY_PAYMENT_SHOPGATE_ORDER_NUMBER_COMMENT_TEXT . " $shopgateOrderNumber\n"
				. "\n", $this->config->getEncoding()
			);
		}
		// Insert transaction number
		$paymentTransactionNumber =
			$oShopgateOrder->getPaymentTransactionNumber();
		if (!empty($paymentTransactionNumber)) {
			$ordersStatusHistory['comments'] .= $this->stringFromUtf8(
				ENTRY_PAYMENT_TRANSACTION_NUMBER_COMMENT_TEXT
				. " $paymentTransactionNumber\n", $this->config->getEncoding()
			);
		}
		// Insert Shipping method
		$shippingInfos = $oShopgateOrder->getShippingInfos();
		if (!empty($shippingInfos)) {
			$ordersStatusHistory['comments'] .= $this->stringFromUtf8(
				ENTRY_SHIPPING_METHOD_COMMENT_TEXT . " "
				. $shippingInfos->getDisplayName() . "("
				. $shippingInfos->getName() . ")\n",
				$this->config->getEncoding()
			);
		}
		// -> Shipping blocked text as the last sentence at the bottom!
		if (!empty($isShippingBlockedComment)) {
			if (!empty($ordersStatusHistory['comments'])) {
				$ordersStatusHistory['comments'] .= "\n";
			}
			$ordersStatusHistory['comments'] .= $this->stringFromUtf8(
				$isShippingBlockedComment['comments'], $this->config->getEncoding()
			);
		}
		$this->insertOrderStatusHistory($ordersStatusHistory);

		// Save payment info
		// -> Output status history comment text and corresponding errors
		if (!empty($statusHistoryComment)) {
			$ordersStatusHistory['comments'] = $this->stringFromUtf8(
				$statusHistoryComment, $this->config->getEncoding()
			);
			$this->insertOrderStatusHistory($ordersStatusHistory);
		}

		// Safe FULL payment info on non generic data (additionally); generic data already has this form of information saved already!
		if (!$isGenericPaymentInfo) {
			$paymentInfo          = $oShopgateOrder->getPaymentInfos();
			$statusHistoryComment = '';

			// Put all available information together
			foreach ($paymentInfo as $key => $value) {
				$statusHistoryComment .= $key . ': ' . $value . "\n";
			}
			$ordersStatusHistory['comments'] = $this->stringFromUtf8(
				$statusHistoryComment, $this->config->getEncoding()
			);
			$this->insertOrderStatusHistory($ordersStatusHistory);
		}

		// order custom fields
		$ordersStatusHistory['comments'] = $this->stringFromUtf8(
			$this->generateCustomFieldsComment(
				$oShopgateOrder, MODULE_PAYMENT_SHOPGATE_ORDER_CUSTOM_FIELDS
			), $this->config->getEncoding()
		);
		if (!empty($ordersStatusHistory['comments'])) {
			$this->insertOrderStatusHistory($ordersStatusHistory);
		}

		$ordersStatusHistory['comments'] = $this->stringFromUtf8(
			$this->generateCustomFieldsComment(
				$oShopgateOrder->getInvoiceAddress(),
				MODULE_PAYMENT_SHOPGATE_INVOICE_ADDRESS_CUSTOM_FIELDS
			), $this->config->getEncoding()
		);
		if (!empty($ordersStatusHistory['comments'])) {
			$this->insertOrderStatusHistory($ordersStatusHistory);
		}

		$ordersStatusHistory['comments'] = $this->stringFromUtf8(
			$this->generateCustomFieldsComment(
				$oShopgateOrder->getDeliveryAddress(),
				MODULE_PAYMENT_SHOPGATE_DELIVERY_ADDRESS_CUSTOM_FIELDS
			), $this->config->getEncoding()
		);
		if (!empty($ordersStatusHistory['comments'])) {
			$this->insertOrderStatusHistory($ordersStatusHistory);
		}

		// Add errors to separate status history entry
		if (!empty($errors)) {
			// Set up errors text
			$ordersStatusHistory['comments'] = ENTRY_ERRORS_EXIST;
			foreach ($errors as $err) {
				$ordersStatusHistory['comments'] .= "\n$err";
			}
			$ordersStatusHistory['comments'] = $this->stringFromUtf8(
				$ordersStatusHistory['comments'], $this->config->getEncoding()
			);
			$this->insertOrderStatusHistory($ordersStatusHistory);
		}

		return true;
	}

	/**
	 * @param $ordersStatusHistory
	 *
	 * @throws ShopgateLibraryException
	 */
	private function insertOrderStatusHistory($ordersStatusHistory)
	{
		// Save entry to DB
		$queryResult = ShopgateWrapper::db_execute_query(
			TABLE_ORDERS_STATUS_HISTORY, $ordersStatusHistory
		);
		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error on inserting an orders status history entry.",
				true
			);
		}
		sleep(1); // Make sure the comments are imported in the same order as they are listed here
	}

	/**
	 * generates a comment from custom fields
	 *
	 * @param ShopgateOrder|ShopgateAddress $object
	 * @param string                        $title
	 *
	 * @return string
	 */
	public function generateCustomFieldsComment($object, $title = "")
	{
		if (empty($object)) {
			return "";
		}

		$comment      = "";
		$customFields = $object->getCustomFields();
		$isFirstItem  = true;
		foreach ($customFields as $customField) {
			$comment .= ((!empty($title) && $isFirstItem) ? "(" . $title . ")" : "")
						. $customField->getLabel() . ": " . $customField->getValue()
						. "\n";
			$isFirstItem = false;
		}

		return $comment;
	}

	/**
	 * Creates a status history entry from one single comment text
	 *
	 * @param        $orderId
	 * @param int    $ordersStatusId
	 * @param string $commentText
	 * @param string $forcedDate
	 *
	 * @return bool
	 * @throws ShopgateLibraryException
	 */
	private function saveSingleOrdersStatusHistory($orderId, $ordersStatusId = 1, $commentText = '', $forcedDate = '')
	{
		if (empty($forcedDate)) {
			$forcedDate = 'now()';
		}
		$ordersStatusHistory = array(
			'orders_id'         => $orderId,
			'orders_status_id'  => $ordersStatusId,
			'date_added'        => $forcedDate,
			'customer_notified' => 0,
			'comments'          => $this->stringFromUtf8($commentText, $this->config->getEncoding())
		);

		$queryResult = ShopgateWrapper::db_execute_query(
			TABLE_ORDERS_STATUS_HISTORY, $ordersStatusHistory
		);
		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error on inserting an orders status history entry.",
				true
			);
		}

		return true;
	}

	/**
	 * @param      $paymentInfos
	 * @param      $dbOrderId
	 * @param      $currentOrderStatus
	 * @param bool $asArray
	 *
	 * @return array|string
	 */
	private function _createPaymentInfos($paymentInfos, $dbOrderId, $currentOrderStatus, $asArray = true)
	{
		$paymentInformation = '';
		foreach ($paymentInfos as $key => $value) {
			$paymentInformation .= $key . ': ' . $value . "\n";
		}

		if ($asArray) {
			return array(
				"orders_id"         => $dbOrderId,
				"orders_status_id"  => $currentOrderStatus,
				"date_added"        => date('Y-m-d H:i:s'),
				"customer_notified" => false,
				"comments"          => ShopgateWrapper::db_prepare_input(
					$paymentInformation
				)
			);
		} else {
			return $paymentInformation;
		}
	}

	/**
	 * Creates a full csv file structure out of the given parameters and possibly some additional helper methods
	 *
	 * @param string $sqlQuery
	 * @param string $subjectName
	 * @param array  $additionalFormatterMethods
	 *
	 * @throws ShopgateLibraryException
	 * @return boolean
	 */
	private function buildCsv($sqlQuery, $subjectName, $additionalFormatterMethods = array())
	{
		if (empty($sqlQuery) || empty($subjectName)) {
			return false;
		}

		// Pluralize the subject name for proper exception output (not correct in some cases like "information", because the plural is the exact same name, a pluralizer-blacklist is needed to handle this)
		$subjectNamePlural = $this->pluralize($subjectName);

		$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);
		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error selecting $subjectNamePlural.", true
			);
		} else {
			// Process each item and put it on the list, that will be exported to the csv-file
			while ($row = ShopgateWrapper::db_fetch_array($queryResult)) {
				// Format the data from raw format to possibly multiple datasets
				$formattedRows = $this->buildSubject(
					$row, $subjectName, $additionalFormatterMethods
				);

				// Skip if there is no data (can happen on failure)
				if ($formattedRows == false) {
					continue;
				}

				// Can contain multiple data arrays
				foreach ($formattedRows as $index => $formattedRow) {
					$this->addItem($formattedRow);
				}
			}
		}

		return true;
	}

	/**
	 * Helper method for valid pluralizing of nouns in most cases (on errors it is possible to add the right plural form as array-entry as the second parameter)
	 *
	 * @param string $subject
	 * @param array  $extendedWordList
	 *
	 * @return string
	 */
	protected function pluralize($subject, $extendedWordList = array())
	{
		$pluralForms = array(// nouns that are already in plural form
							 'information' => 'information',
							 'sheep'       => 'sheep',
							 'fish'        => 'fish',
							 'hair'        => 'hair',
							 'fruit'       => 'fruit',
							 'aircraft'    => 'aircraft',
							 'chinese'     => 'chinese',
							 'japanese'    => 'japanese',
							 'swiss'       => 'swiss',
							 // irregular nouns
							 'child'       => 'children',
							 'ox'          => 'oxen',
							 'man'         => 'men',
							 'woman'       => 'women',
							 'foot'        => 'feet',
							 'goose'       => 'geese',
							 'tooth'       => 'teeth',
							 'mouse'       => 'mice',
							 'louse'       => 'lice',
							 // o at end for living things get es appended, otherwise just s
							 'hero'        => 'heroes',
							 'potato'      => 'potatoes',
							 'tomato'      => 'tomatoes',// possibly more
		);

		if (!empty($extendedWordList)) {
			foreach ($extendedWordList as $key => $value) {
				// key and value must both exist
				$key   = strtolower(trim($key));
				$value = strtolower(trim($value));

				// non-associative (indexed) arrays must have the keys deleted
				if (is_numeric($key)) {
					$key = '';
				}

				// ignore fully empty fields and copy non-empty keys to empty values and backwards
				if (empty($key) && empty($value)) {
					continue;
				} else {
					if (empty($key)) {
						$key = $value;
					} else {
						if (empty($value)) {
							$value = $key;
						}
					}
				}

				// append all extended words in lowercase
				$pluralForms[$key] = $value;
			}
		}

		// Check whitelist with predefined plural forms first (keep case of the first letter)
		$pluralized = trim($subject);
		if (isset($pluralForms[strtolower($pluralized)])) {
			return substr($pluralized, 0, 1) . substr(
				$pluralForms[strtolower($pluralized)], 1
			);
		}

		$endingChar = strtolower(substr($pluralized, -1));
		$prevChar   = strtolower(substr($pluralized, -2, -1));
		if ($endingChar != 's' && $prevChar != 's') { // not ss at end
			if ($endingChar == 'y') {
				$pluralized = substr($pluralized, 0, -1) . 'ies';
			} else {
				if ($prevChar != 'f' && $endingChar == 'f') { // ff at end
					$pluralized = substr($pluralized, 0, -1) . 'ves';
				} else {
					if ($prevChar == 'f' && $endingChar == 'e') { // fe at end
						$pluralized = substr($pluralized, 0, -2) . 'ves';
					} else {
						if ($prevChar == 's' && $endingChar == 'h'
							|| $prevChar == 'c' && $endingChar == 'h'
							|| $endingChar == 'x'
							|| $prevChar == 's' && $endingChar == 's'
						) { // ch or sh or x or ss
							$pluralized .= 'es';
						} else {
							$pluralized .= 's';
						}
					}
				}
			}
		}

		return $pluralized;
	}

	/**
	 * Takes the given data and maps it to the internal structure using the corresponding helper methods, called by the "executeLoaders" method
	 * and by the additional modificator (formatter) methods
	 *
	 * @param array  $rawSubjectData
	 * @param string $subjectName
	 * @param array  $additionalFormatterMethods
	 *
	 * @return two-dimensional:array
	 */
	private function buildSubject($rawSubjectData, $subjectName, $additionalFormatterMethods = array())
	{
		$formattedSubjectData = array();

		$subjectName = trim($subjectName);
		if (empty($subjectName)) {
			return false;
		}

		// Build the names for the methods to be called
		$subjectName         = strtolower($subjectName);
		$buildDefaultRow     = 'buildDefault' . ucfirst($subjectName) . 'Row';
		$getCreateCsvLoaders = "getCreate" . ucfirst(
				$this->camelize($this->pluralize($subjectName))
			) . "CsvLoaders";

		// Build a default array structure where the formatted data will be stored in by calling the specific class-method (by name -> the method must exist)
		if (!method_exists($this, $buildDefaultRow)) {
			return false;
		}
		// -> get only one row, so the index 0 is used
		$formattedSubjectData[0] = $this->{$buildDefaultRow}();

		// Get a list of all processing methods by calling the specific class-method (by name -> the method must exist)
		if (!method_exists($this, $getCreateCsvLoaders)) {
			return false;
		}
		$csvLoaders = $this->{$getCreateCsvLoaders}();

		// Map the raw data to the internal structure provided before
		// -> there is still only one row there
		$formattedSubjectData[0] = $this->executeLoaders(
			$csvLoaders, $formattedSubjectData[0], $rawSubjectData
		);

		// There can be more methods given to modify (or fill) the corresponding subject (i.e options or attributes for products)
		// -> the additional formatter methods can append additional rows!
		foreach ($additionalFormatterMethods as $formatMethod) {
			if (!empty($formatMethod)) {
				if (method_exists($this, $formatMethod)) {
					$formattedSubjectData = $this->{$formatMethod}(
						$formattedSubjectData, $rawSubjectData
					);
				}
			}
		}

		// Returns all fully formatted data rows for the subject (can be multiple rows after post processing, using the additional formatter methods)
		return $formattedSubjectData;
	}

	/**
	 * @return array
	 */
	public function _getPluginInfoLoaders()
	{
		return array('_loadPluginInfo_name', '_loadPluginInfo_version', '_loadPluginInfo_edition');
	}

	/**
	 * tax classes
	 *
	 * @return array
	 * @throws ShopgateLibraryException
	 */
	protected function getTaxClasses()
	{
		$sqlQuery = "SELECT `tbl_tc`.`tax_class_id` 'id', `tbl_tc`.`tax_class_title` 'key' FROM `" . TABLE_TAX_CLASS
					. "` as tbl_tc";

		$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);
		$result      = array();
		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error selecting.", true
			);
		} else {
			while ($row = ShopgateWrapper::db_fetch_array($queryResult)) {
				array_push($result, $row);
			}
		}

		return $result;
	}

	/**
	 * get tax rates
	 *
	 * @return array<string, mixed>[]
	 * @throws ShopgateLibraryException
	 */
	protected function getTaxRates()
	{
		$query = "SELECT tr.tax_rates_id, tr.tax_description, tr.tax_rate, tr.tax_priority, "
				 . "c.countries_iso_code_2, z.zone_code, tc.tax_class_id, tc.tax_class_title "
				 . "FROM `" . TABLE_TAX_RATES . "` AS tr "
				 . "JOIN `" . TABLE_GEO_ZONES . "` AS gz ON tr.tax_zone_id = gz.geo_zone_id "
				 . "JOIN `" . TABLE_ZONES_TO_GEO_ZONES . "` AS ztgz ON gz.geo_zone_id = ztgz.geo_zone_id "
				 . "JOIN `" . TABLE_COUNTRIES . "` AS c ON ztgz.zone_country_id = c.countries_id "
				 . "LEFT OUTER JOIN `" . TABLE_ZONES . "` AS z ON ztgz.zone_id = z.zone_id "
				 . // zone (aka state) might not be mapped, rate applies for whole country in that case
				 "JOIN `" . TABLE_TAX_CLASS . "` tc ON tr.tax_class_id = tc.tax_class_id;";

		$result = ShopgateWrapper::db_query($query);
		if (!$result) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error selecting.", true
			);
		}

		$taxRates = array();
		while ($row = ShopgateWrapper::db_fetch_array($result)) {
			$taxRates[] = $row;
		}

		return $taxRates;
	}

	/**
	 * @param ShopgateCart $cart
	 *
	 * @return array
	 */
	private function checkCoupon(ShopgateCart $cart)
	{
		//check if ot_discount module is installed 
		if (defined("MODULE_ORDER_TOTAL_INSTALLED")
			&& strpos(MODULE_ORDER_TOTAL_INSTALLED, "ot_discount_coupon.php") !== false
		) {

			$file = (rtrim(DIR_FS_CATALOG, "/") . "/" . rtrim(DIR_WS_CLASSES, "/") . "/discount_coupon.php");

			if (!file_exists($file)) {
				return array();
			}

			include_once($file);

			$coupons              = $cart->getExternalCoupons();
			$productAmountWithTax = $this->cartItemModel->getProductAmountWithTax($cart);
			$hasValidCoupon       = false;

			foreach ($coupons as &$coupon) {
				$code       = $coupon->getCode();
				$osCoupon   = new discount_coupon($code, $this->couponModel->getAddress($cart));
				$shopCoupon = $osCoupon->coupon;

				$coupon->setCurrency($this->config->getCurrency());
				$coupon->setName($shopCoupon["coupons_description"]);

				if (empty($customer_id)) {
					$coupon->setIsValid(false);
					$coupon->setNotValidMessage("Customer not registered in shop system");
					continue;
				}

				/* Coupon exist */
				if (empty($shopCoupon)) {
					$coupon->setIsValid(false);
					$coupon->setNotValidMessage("Coupon not found");
					continue;
				}

				/* Coupon in time windows */
				if (!$this->couponModel->isCouponInsideOfATimeFrame($shopCoupon)) {
					$coupon->setIsValid(false);
					$coupon->setNotValidMessage("Coupon out of date");
					continue;
				}

				$minOrderType = $shopCoupon["coupons_min_order_type"];

				if ($minOrderType == "price") {
					if ($productAmountWithTax < $shopCoupon["coupons_min_order"]) {
						$coupon->setIsValid(false);
						$coupon->setNotValidMessage(
							"Minimum order value not reached"
						);
					} else {
						$coupon->setIsValid(true);
					}
				} elseif ($minOrderType == "quantity") {
					if (count($cart->getItems()) < $shopCoupon["coupons_min_order"]) {
						$coupon->setIsValid(false);
						$coupon->setNotValidMessage("Coupon out of date");
						continue;
					} else {
						$coupon->setIsValid(true);
					}
				}

				switch ($shopCoupon["coupons_discount_type"]) {
					case "percent":
						$coupon->setAmountGross($productAmountWithTax * $shopCoupon["coupons_discount_amount"]);
						break;
					case "fixed":
						$coupon->setAmountGross($shopCoupon["coupons_discount_amount"]);
						break;
					case "shipping":
						$coupon->setAmountGross($cart->getAmountShipping() * $shopCoupon["coupons_discount_amount"]);
						break;
					default:
						break;
				}

				$products = $cart->getItems();

				if (count($products) == 0) {
					$coupon->setIsValid(false);
					$coupon->setNotValidMessage("No products");
					break;
				}

				foreach ($products as $product) {
					$pId = $product->getItemNumber();

					/* excluded categories	*/
					if ($this->couponModel->isProductExcludedFromCategory($pId, $code)) {
						$coupon->setIsValid(false);
						$coupon->setNotValidMessage("Some or all of the products are in an excluded category");
						continue(2);
					}

					/* excluded products */
					if ($this->couponModel->isProductExcluded($pId, $code)) {
						$coupon->setIsValid(false);
						$coupon->setNotValidMessage(
							"Some or all of the products in your cart are excluded from this coupon"
						);
						continue(2);
					}

					/*	excluded manufacturers	*/
					if ($this->couponModel->isManufacturerExcluded($pId, $code)) {
						$coupon->setIsValid(false);
						$coupon->setNotValidMessage("Some or all of the manufacturers in your cart are excluded");
						continue(2);
					}

				} // EOF product iteration

				/* excluded customers*/
				if ($this->couponModel->isCustomerExcluded($code, $customer_id)) {
					$coupon->setIsValid(false);
					$coupon->setNotValidMessage("customer is excluded for this coupon");
					continue;
				}

				/* customer never ordered in this shop */
				if ($this->couponModel->customerHasOrders($code, $customer_id)) {
					$coupon->setIsValid(false);
					$coupon->setNotValidMessage("Customer never ordered something in this shop.");
					continue;
				}

				if ($coupon->getIsValid() && !$hasValidCoupon) {
					$hasValidCoupon = true;
				} elseif ($coupon->getIsValid() && $hasValidCoupon) {
					$coupon->setIsValid(false);
					$coupon->setNotValidMessage("This coupon code can\'t be used with others at the same time.");
				}
			}

			return $coupons;
		}
	}

	/**
	 * return an array with all valid shipping methods to an order
	 *
	 * @param ShopgateCart $sgShoppingCart
	 *
	 * @return array of ShopgateShippingMethod
	 */
	private function getShipping(ShopgateCart $sgShoppingCart)
	{
		$resultShippingMethods = array();

		if (!defined('MODULE_SHIPPING_INSTALLED') || !tep_not_null(MODULE_SHIPPING_INSTALLED)
			|| MODULE_SHIPPING_INSTALLED == ""
		) {
			return array();
		}

		global $total_count, $shipping_weight, $total_weight, $shipping_num_boxes;

		$total_count        = count($sgShoppingCart->getItems());
		$total_weight       = $this->cartItemModel->getProductsWeight(
			$sgShoppingCart->getItems()
		);
		$shipping_num_boxes = 1;

		include_once(rtrim(DIR_WS_CLASSES, "/") . "/currencies.php");
		global $currencies;
		$currencies = new currencies();

		include_once(rtrim(DIR_WS_CLASSES, "/") . "/shopping_cart.php");
		global $cart;
		$cart = new shoppingCart();

		foreach ($sgShoppingCart->getItems() as $product) {
			$options   = $product->getOptions();
			$sgOptions = array();

			foreach ($options as $option) {
				$sgOptions[$option->getOptionNumber()] = $option->getValueNumber();
			}
			$cart->add_cart($product->getItemNumber(), $product->getQuantity(), $sgOptions);
		}

		include_once(rtrim(DIR_WS_CLASSES, "/") . "/order.php");
		global $order, $sendto, $billto;

		$sgDeliverAddress = $sgShoppingCart->getDeliveryAddress();

		if (!empty($sgDeliverAddress)) {
			$country = $this->locationModel->getCountryByIso2Name($sgDeliverAddress->getCountry());
			$zone    = $this->locationModel->getZoneByCountryIdModuleShippingFlat($country["countries_id"]);

			$sendto = array(
				"firstname"          => $sgDeliverAddress->getFirstName(),
				"lastname"           => $sgDeliverAddress->getLastName(),
				"company"            => $sgDeliverAddress->getCompany(),
				"street_address"     => $sgDeliverAddress->getStreet1(),
				"suburb"             => "",
				"postcode"           => $sgDeliverAddress->getZipcode(),
				"city"               => $sgDeliverAddress->getCity(),
				"zone_id"            => $zone["zone_id"],
				"zone_name"          => $zone["zone_name"],
				"country_id"         => $country["countries_id"],
				"country_iso_code_2" => $country["countries_iso_code_2"],
				"country" => array("iso_code_2" => $country["countries_iso_code_2"]),
				"country_iso_code_3" => $country["countries_iso_code_3"],
				"address_format_id"  => ""
			);
		}

		$sgInvoiceAddress = $sgShoppingCart->getInvoiceAddress();

		if (empty($sgInvoiceAddress)) {
			$billto = $sendto;
		} else {
			$country = $this->locationModel->getCountryByIso2Name($sgInvoiceAddress->getCountry());
			$zone    = $this->locationModel->getZoneByCountryId($country["countries_id"]);
			$billto  = array(
				"firstname"          => $sgInvoiceAddress->getFirstName(),
				"lastname"           => $sgInvoiceAddress->getLastName(),
				"company"            => $sgInvoiceAddress->getCompany(),
				"street_address"     => $sgInvoiceAddress->getStreet1(),
				"suburb"             => "",
				"postcode"           => $sgInvoiceAddress->getZipcode(),
				"city"               => $sgInvoiceAddress->getCity(),
				"zone_id"            => $zone["zone_id"],
				"zone_name"          => $zone["zone_name"],
				"country_id"         => $country["countries_id"],
				"country_iso_code_2" => $country["countries_iso_code_2"],
				"country" => array("iso_code_2" => $country["countries_iso_code_2"]),
				"country_iso_code_3" => $country["countries_iso_code_3"],
				"address_format_id"  => ""
			);
		}
		$order = new order();
		$order->cart();
		$order->customer = $sendto;
		$order->delivery = $sendto;
		$order->billing  = $billto;

		include_once(rtrim(DIR_WS_MODULES, "/") . "/order_total/ot_shipping.php");
		global $ot_shipping;
		$ot_shipping = new ot_shipping();
		$ot_shipping->process();

		require(DIR_WS_CLASSES . 'shipping.php');
		$shipping_modules = new shipping;

		if (defined('MODULE_ORDER_TOTAL_SHIPPING_FREE_SHIPPING')
			&& (MODULE_ORDER_TOTAL_SHIPPING_FREE_SHIPPING == 'true')
		) {
			$pass = false;

			switch (MODULE_ORDER_TOTAL_SHIPPING_DESTINATION) {
				case 'national':
					if ($order->delivery['country_id'] == STORE_COUNTRY) {
						$pass = true;
					}
					break;
				case 'international':
					if ($order->delivery['country_id'] != STORE_COUNTRY) {
						$pass = true;
					}
					break;
				case 'both':
					$pass = true;
					break;
			}

			$free_shipping = false;
			if (($pass == true) && ($order->info['total'] >= MODULE_ORDER_TOTAL_SHIPPING_FREE_SHIPPING_OVER)) {
				$free_shipping = true;
				include(DIR_WS_LANGUAGES . $this->language . '/modules/order_total/ot_shipping.php');
			}
		} else {
			$free_shipping = false;
		}

		// if shipping is free all other shipping methods will be ignored
		if ($free_shipping) {
			$sgShippingMethod = new ShopgateShippingMethod();
			$sgShippingMethod->setDescription(
				"Total amount over "
				. MODULE_ORDER_TOTAL_SHIPPING_FREE_SHIPPING_OVER . " is free"
			);
			$sgShippingMethod->setTitle("Free Shipping");
			$sgShippingMethod->setAmount(0);
			return array($sgShippingMethod);
		}

		$quotes = $shipping_modules->quote();

		foreach ($quotes AS $shippingModule) {
			//we don`t support usps as shopgate plugin shipping method
			//also on error continue
			if (strpos($shippingModule['module'], "United States Postal Service") !== false
				|| !empty($shippingModule['error'])
			) {
				continue;
			}

			$sgShippingMethod = new ShopgateShippingMethod();
			$sgShippingMethod->setId($shippingModule["id"]);
			$sgShippingMethod->setTitle($shippingModule["module"]);

			if (!empty($shippingModule["methods"]) && is_array($shippingModule["methods"])) {
				foreach ($shippingModule["methods"] as $method) {
					$tmp_shipping = $sgShippingMethod;
					$cost         = $method["cost"];
					if (isset($shippingModule["tax"]) && !empty($shippingModule["tax"])) {
						$tmp_shipping->setTaxClass($this->taxModel->getTaxClassByValue($shippingModule["tax"]));
						$tmp_shipping->setTaxPercent($shippingModule['tax']);
						$costWithTax = $this->formatPriceNumber($cost * (1 + ($shippingModule["tax"] / 100)), 2);
						$tmp_shipping->setAmountWithTax($costWithTax);
						$tmp_shipping->setAmount($cost);
					} else {
						$tmp_shipping->setTaxPercent(0);
						$tmp_shipping->setAmountWithTax($cost);
						$tmp_shipping->setAmount($cost);
					}

					$resultShippingMethods[] = $tmp_shipping;
				}
			}
		}

		return $resultShippingMethods;
	}

	/**
	 * @param ShopgateCart $cart
	 *
	 * @return array
	 */
	private function checkValidityOfCartItems(ShopgateCart $cart)
	{
		$resultItems = array();

		foreach ($cart->getItems() as $item) {

			$pId       = $this->cartItemModel->getProductIdFromCartItems($item);
			$sgProduct = new ShopgateCartItem();

			if (!$this->itemModel->productExist($pId)) {
				$sgProduct->setError(ShopgateLibraryException::CART_ITEM_PRODUCT_NOT_FOUND);
				$sgProduct->setIsBuyable(false);
				$resultItems[] = $sgProduct;
				continue;
			} else {
				global $customer_zone_id, $customer_country_id;
				$address = $cart->getDeliveryAddress();

				if (!empty($address)) {
					$country             = $this->locationModel->getCountryByIso2Name($address->getCountry());
					$customer_country_id = $country["countries_id"];
					$customer_zone_id    = $this->locationModel->getCustomerZoneId(
						$customer_country_id,
						ShopgateMapper::getShoppingsystemStateCode($address->getState())
					);
					$productTaxRate      = tep_get_tax_rate(
						$this->taxModel->getProductsTaxClass($pId),
						$customer_country_id, $customer_zone_id
					);
				} else {
					$productTaxRate = tep_get_tax_rate($this->taxModel->getProductsTaxClass($pId));
				}

				$sgProduct->setIsBuyable(true);
				$sgProduct = $this->cartItemModel->generateCartItemProduct($item, $sgProduct);
				$sgProduct = $this->cartItemModel->getCartItemOptions($item, $sgProduct, $productTaxRate);
				$sgProduct = $this->cartItemModel->getCartItemAttributes($item, $sgProduct);
				$dbProduct = $this->itemModel->getProduct($pId);

				$tmpPriceItem = array(
					"specials_new_products_price" => tep_get_products_special_price($pId),
					"products_price"              => $dbProduct["products_price"]
				);

				$price = $this->getPrice($tmpPriceItem, $productTaxRate, $this->exchangeRate);
				$sgProduct->setUnitAmountWithTax($price);
				$sgProduct->setUnitAmount(
					(empty($productTaxRate)) ? $price : $this->formatPriceNumber($price / (1 + ($productTaxRate / 100)))
				);
				$sgProduct->setQtyBuyable($dbProduct["products_quantity"]);
				$sgProduct->setStockQuantity($dbProduct["products_quantity"]);

				if (empty($dbProduct["products_quantity"]) || $dbProduct["products_quantity"] == 0) {
					$sgProduct->setIsBuyable(false);
					$sgProduct->setError(ShopgateLibraryException::CART_ITEM_PRODUCT_NOT_FOUND);
				}

				if ($dbProduct["products_status"] == 0) {
					$sgProduct->setIsBuyable(false);
					$sgProduct->setError(
						ShopgateLibraryException::CART_ITEM_PRODUCT_NOT_FOUND
					);
				}

				$resultItems[] = $sgProduct;
			}
		}
		return $resultItems;
	}

	/**
	 * Checks if there is a special price and returns the special price with tax, if given and adds up exchange if given OR
	 * the default price if no special is set
	 *
	 * @param array  $aArticle
	 * @param string $taxPercent
	 * @param int    $exchangeValue
	 *
	 * @return string - empty if there is no special price or a float value with the old price [with tax and exchange]
	 */
	protected function getPrice($aArticle, $taxPercent = null, $exchangeValue = 1)
	{
		if ($this->isSpecialPrice($aArticle)) {
			$price = $aArticle["specials_new_products_price"];
		} else {
			$price = $aArticle["products_price"];
		}

		if (!empty($price)) {
			if (!empty($taxPercent)) {
				$price *= 1 + ($taxPercent / 100);
			}
			if (!empty($exchangeValue)) {
				$price *= $exchangeValue;
			}

			$price = $this->formatPriceNumber($price, 2);
		}

		return $price;
	}

	/**
	 * Checks if the product has a valid special price (returns true if so, otherwise it returns false)
	 *
	 * @param array $aArticle
	 *
	 * @return boolean
	 */
	protected function isSpecialPrice($aArticle)
	{
		return !empty($aArticle["specials_new_products_price"]);
	}

	/**
	 * Takes all as "shipping completed" marked orders and apply this status to the equivalent Shopgate order.
	 *
	 * @param $errCount
	 * @param $message
	 *
	 * @return bool
	 * @throws ShopgateLibraryException
	 */
	protected function cronSetOrdersShippingCompleted(&$errCount, &$message)
	{
		if (!$this->tableExists(TABLE_ORDERS_SHOPGATE_ORDER)) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error: \"" . TABLE_ORDERS_SHOPGATE_ORDER
				. "\" table is missing.", true
			);
		}
		$query  = "SELECT `sgo`.`orders_id`, `sgo`.`shopgate_order_number` "
				  . "FROM `" . TABLE_ORDERS_SHOPGATE_ORDER . "` sgo "
				  . "INNER JOIN `" . TABLE_ORDERS . "` osco ON (`osco`.`orders_id` = `sgo`.`orders_id`) "
				  . "WHERE `sgo`.`is_sent_to_shopgate` = 0 " . "AND ("
				  . "`sgo`.`shopgate_shop_number` = '" . ShopgateWrapper::db_input($this->config->getShopNumber())
				  . "' " . "OR `sgo`.`shopgate_shop_number` IS NULL"
				  . ") " . "AND `osco`.`orders_status` = '"
				  . ShopgateWrapper::db_input($this->config->getOrderStatusShipped()) . "';";
		
		$result = ShopgateWrapper::db_query_print_on_err($query);

		if (empty($result)) {
			return true;
		}
		while ($shopgateOrder = ShopgateWrapper::db_fetch_array($result)) {
			if (!$this->setOrderShippingCompleted(
				$shopgateOrder, $shopgateOrder['orders_id'], $this->merchantApi,
				$this->config
			)
			) {
				$errCount++;
				$message .= 'Shopgate order number "'
							. $shopgateOrder['shopgate_order_number'] . '": error'
							. "\n";
			}
		}
	}

	/**
	 * request Shopgate to set shipping complete to an order
	 *
	 * @param                          string [} $shopgateOrder
	 * @param                          $orderId
	 * @param ShopgateMerchantApi      $merchantApi
	 * @param ShopgateConfigOsCommerce $config
	 *
	 * @return bool
	 * @throws ShopgateLibraryException
	 */
	protected function setOrderShippingCompleted(
		$shopgateOrder, $orderId, ShopgateMerchantApi &$merchantApi,
		ShopgateConfigOsCommerce &$config
	) {
		$success = false;

		// These are expected and should not be added to error count:
		$ignoreCodes = array(
			ShopgateMerchantApiException::ORDER_ALREADY_COMPLETED,
			ShopgateMerchantApiException::ORDER_SHIPPING_STATUS_ALREADY_COMPLETED
		);

		try {
			$merchantApi->setOrderShippingCompleted($shopgateOrder['shopgate_order_number']);
			$this->saveSingleOrdersStatusHistory(
				$orderId, $config->getOrderStatusShipped(),
				"[Shopgate] " . ENTRY_ORDER_MARKED_AS_SHIPPED
			);
			$success = true;
		} catch (ShopgateLibraryException $e) {
			$response = $this->stringFromUtf8($e->getAdditionalInformation(), $config->getEncoding());
			$this->saveSingleOrdersStatusHistory(
				$orderId, $config->getOrderStatusShipped(),
				"[Shopgate] " . ENTRY_SHOPGATE_MODULE_ERROR . " ({$e->getCode()}): {$response}"
			);
		} catch (ShopgateMerchantApiException $e) {
			$response = $this->stringFromUtf8($e->getMessage(), $config->getEncoding());
			$this->saveSingleOrdersStatusHistory(
				$orderId, $config->getOrderStatusShipped(),
				"[Shopgate] " . ENTRY_SHOPGATE_MODULE_ERROR . " ({$e->getCode()}): {$response}"
			);
			$success = (in_array($e->getCode(), $ignoreCodes)) ? true : false;
		} catch (Exception $e) {
			$response = $this->stringFromUtf8($e->getMessage(), $config->getEncoding());
			$this->saveSingleOrdersStatusHistory(
				$orderId, $config->getOrderStatusShipped(),
				"[Shopgate] " . ENTRY_SHOPGATE_UNKNOWN_ERROR . " ({$e->getCode()}): {$response}"
			);
		}

		// Update shopgate order on success
		if ($success) {
			$qry = 'UPDATE `' . TABLE_ORDERS_SHOPGATE_ORDER
				   . '` SET `is_sent_to_shopgate` = 1 WHERE `shopgate_order_number` = '
				   . $shopgateOrder['shopgate_order_number'] . ';';
			ShopgateWrapper::db_query($qry);
		}

		return $success;
	}

	/**
	 * logic for cron job CANCEL_ORDERS.
	 * Takes all as "cancel" marked orders and apply this status to the equivalent Shopgate order.
	 *
	 * @param $message
	 * @param $errCount
	 *
	 * @throws ShopgateLibraryException
	 */
	protected function cronCancelOrder(&$message, &$errCount)
	{
		if (!$this->tableExists(TABLE_ORDERS_SHOPGATE_ORDER)) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error: \"" . TABLE_ORDERS_SHOPGATE_ORDER
				. "\" table is missing.", true
			);
		}

		$query  = "SELECT `sgo`.`orders_id`, `sgo`.`shopgate_order_number`, `sgo`.`order_data` "
				  . "FROM `" . TABLE_ORDERS_SHOPGATE_ORDER . "` sgo "
				  . "INNER JOIN `" . TABLE_ORDERS . "` osco ON (`osco`.`orders_id` = `sgo`.`orders_id`) "
				  . "WHERE `sgo`.`is_cancellation_sent_to_shopgate` = 0 "
				  . "AND (" . "`sgo`.`shopgate_shop_number` = '"
				  . ShopgateWrapper::db_input($this->config->getShopNumber()) . "' "
				  . "OR `sgo`.`shopgate_shop_number` IS NULL"
				  . ") " . "AND `osco`.`orders_status` = '"
				  . ShopgateWrapper::db_input($this->config->getOrderStatusCanceled()) . "';";
		$result = ShopgateWrapper::db_query_print_on_err($query);

		if (empty($result)) {
			return;
		}
		while ($shopgateOrder = ShopgateWrapper::db_fetch_array($result)) {
			if (!$this->setOrderCanceled(
				$shopgateOrder, $shopgateOrder['orders_id'], $this->merchantApi, $this->config
			)
			) {
				$errCount++;
				$message .= 'Shopgate order number "' . $shopgateOrder['shopgate_order_number'] . '": error' . "\n";
			}
		}
	}

	/**
	 * request Shopgate to cancel/cancel parts(shipping) of an order
	 *
	 * @param                          $shopgateOrder
	 * @param                          $orderId
	 * @param ShopgateMerchantApi      $merchantApi
	 * @param ShopgateConfigOsCommerce $config
	 *
	 * @return bool
	 * @throws ShopgateLibraryException
	 */
	protected function setOrderCanceled(
		$shopgateOrder, $orderId, ShopgateMerchantApi &$merchantApi,
		ShopgateConfigOsCommerce &$config
	) {

		$success = false;
		$ignoreCodes
				 = array( // These are expected and should not be added to error count:
						  222, //Order already cancelled
		);

		try {
			$orderStatus       = $config->getOrderStatusCanceled();
			$cancellationItems = array();

			// fallback when order data was not inserted into db. 
			if (empty($shopgateOrder) || empty($shopgateOrder["order_data"])) {
				$params = array(
					'shop_number'   => $this->config->getShopNumber(), 'with_items' => 1,
					'order_numbers' => array(0 => "{$shopgateOrder['shopgate_order_number']}"),
				);

				/* @var ShopgateMerchantApiResponse $order */
				$merchantOrderResponse = $merchantApi->getOrders($params);
				$sgRequestedOrders     = $merchantOrderResponse->getData();
				if (count($sgRequestedOrders) == 0) {
					$this->saveSingleOrdersSatusHistory(
						$shopgateOrder["orders_id"], $orderStatus,
						"[Shopgate] " . SHOPGATE_ORDER_NOT_FOUND
					);
					return false;
				}

				/* @var ShopgateOrder $shopgateOrderObject */
				$shopgateOrderObject = reset(
					$sgRequestedOrders
				);// there can be only one order. cause only one was requested

			} else {
				$shopgateOrderObject = unserialize(
					$shopgateOrder["order_data"]
				);
			}

			/* @var ShopgateOrder $shopgateOrderObject */
			if (!$config->getAlwaysCancelShipping()) {
				foreach ($shopgateOrderObject->getItems() as $item) {
					$cancellationItems[]
						= array(
						'order_item_id' => $item->getOrderItemId(),
						'item_number'   => (!is_null($item->getParentItemNumber())
							? $item->getParentItemNumber()
							: $item->getItemNumber()),
						'quantity'      => $item->getQuantity(),
					);
				}
			}

			// if cancelShipping is true, cancelCompleteOrder must be true too
			$merchantApi->cancelOrder(
				$shopgateOrder['shopgate_order_number'],
				$config->getAlwaysCancelShipping(), $cancellationItems,
				$config->getAlwaysCancelShipping()
			);
			$this->saveSingleOrdersSatusHistory(
				$shopgateOrder["orders_id"], $orderStatus,
				"[Shopgate] " . ENTRY_ORDER_MARKED_AS_CANCELED
			);
			$success = true;
		} catch (ShopgateLibraryException $e) {
			$response = $this->stringFromUtf8(
				$e->getAdditionalInformation(), $config->getEncoding()
			);
			$this->saveSingleOrdersSatusHistory(
				$orderId, $orderStatus,
				"[Shopgate] " . ENTRY_SHOPGATE_MODULE_ERROR . " ({$e->getCode()}): {$response}"
			);
		} catch (ShopgateMerchantApiException $e) {
			$response = $this->stringFromUtf8($e->getMessage(), $config->getEncoding());
			$this->saveSingleOrdersSatusHistory(
				$orderId, $orderStatus,
				"[Shopgate] " . ENTRY_SHOPGATE_MODULE_ERROR . " ({$e->getCode()}): {$response}"
			);
			$success = (in_array($e->getCode(), $ignoreCodes)) ? true : false;
		} catch (Exception $e) {
			$response = $this->stringFromUtf8($e->getMessage(), $config->getEncoding());
			$this->saveSingleOrdersSatusHistory(
				$orderId, $orderStatus,
				"[Shopgate] " . ENTRY_SHOPGATE_UNKNOWN_ERROR . " ({$e->getCode()}): {$response}"
			);
		}

		// Update shopgate order on success
		if ($success) {
			$qry = 'UPDATE `' . TABLE_ORDERS_SHOPGATE_ORDER
				   . '` SET `is_cancellation_sent_to_shopgate` = 1 WHERE `shopgate_order_number` = '
				   . $shopgateOrder['shopgate_order_number'] . ';';
			ShopgateWrapper::db_query($qry);
		}

		return $success;
	}

	/**
	 * @param $aInfo
	 * @param $unused
	 *
	 * @return mixed
	 */
	public function _loadPluginInfo_name($aInfo, $unused)
	{
		$aInfo['system_name'] = (!empty($this->swisscartVersion) ? 'swisscart / osCommerce' : 'osCommerce');

		return $aInfo;
	}

	/**
	 * @param $aInfo
	 * @param $unused
	 *
	 * @return mixed
	 */
	public function _loadPluginInfo_version($aInfo, $unused)
	{
		$aInfo['version'] = defined('PROJECT_VERSION') ? PROJECT_VERSION : 'undefined';

		return $aInfo;
	}

	/**
	 * @param $aInfo
	 * @param $unused
	 *
	 * @return mixed
	 */
	public function _loadPluginInfo_edition($aInfo, $unused)
	{
		$aInfo['edition'] = (!empty($this->swisscartVersion) ? $this->swisscartVersion : '');

		return $aInfo;
	}

	/**
	 * @param $orderIds
	 * @param $status
	 *
	 * @throws ShopgateLibraryException
	 */
	public function updateOrdersStatus($orderIds, $status)
	{
		if (!is_array($orderIds) || empty($orderIds)) {
			return;
		}

		// Get shopgate orders
		if (!$this->tableExists(TABLE_ORDERS_SHOPGATE_ORDER)) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error: \"" . TABLE_ORDERS_SHOPGATE_ORDER
				. "\" table is missing.", true
			);
		}

		$query  = ShopgateWrapper::db_input(
			"SELECT `sgo`.`orders_id`, `sgo`.`shopgate_order_number`, `sgo`.`shopgate_shop_number` "
			. "FROM `" . TABLE_ORDERS_SHOPGATE_ORDER . "` sgo "
			. "WHERE `sgo`.`orders_id` IN (" . ShopgateWrapper::db_input(
				implode(", ", $orderIds)
			) . ")"
		);
		$result = ShopgateWrapper::db_query($query);

		if (empty($result)) {
			return;
		}
		$configurations = array();
		while ($shopgateOrder = ShopgateWrapper::db_fetch_array($result)) {
			$shopNumber = !empty($shopgateOrder['shopgate_shop_number'])
				? $shopgateOrder['shopgate_shop_number'] : 'global';

			if (empty($merchantApis[$shopNumber])) {
				try {
					$config = new ShopgateConfigOsCommerce();

					if ($shopNumber != 'global') {
						$config->loadByShopNumber($shopNumber);
					}

					$builder                     = new ShopgateBuilder($config);
					$merchantApis[$shopNumber]   = $builder->buildMerchantApi();
					$configurations[$shopNumber] = $config;
				} catch (ShopgateLibraryException $e) {
					$this->log(
						'error in update orders status: ' . $e,
						ShopgateLogger::LOGTYPE_ERROR
					);
				}

				if ($status == $configurations[$shopNumber]->getOrderStatusCanceled()) {
					$this->setOrderCanceled(
						$shopgateOrder, $shopgateOrder['orders_id'],
						$merchantApis[$shopNumber], $configurations[$shopNumber]
					);
				} elseif ($status == $configurations[$shopNumber]->getOrderStatusShipped()) {
					$this->setOrderShippingCompleted(
						$shopgateOrder, $shopgateOrder['orders_id'],
						$merchantApis[$shopNumber], $configurations[$shopNumber]
					);
				}
			}
		}
	}

	###########################################################################
	## ITEM EXPORT METHODS                                                   ##
	###########################################################################

	/**
	 * @param int|null $limit
	 * @param int|null $offset
	 * @param array    $uids
	 *
	 * @throws ShopgateLibraryException
	 */
	protected function createItems($limit = null, $offset = null, array $uids = array())
	{
		$this->itemModel->setLanguageId($this->langID);
		$this->itemModel->setTaxZone($this->taxZone);
		$this->itemModel->setIsUsaPlugin($this->config->getIsUsaPlugin());
		$sqlQuery    = $this->itemModel->buildItemQuery($limit, $offset, $uids);
		$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);

		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error selecting Items.", true
			);
		} else {
			while ($row = ShopgateWrapper::db_fetch_array($queryResult)) {
				$itemModel = new ShopgateItemXmlModel();
				$itemModel->setItem($row);
				$itemModel->setLanguageId($this->langID);
				$itemModel->setStoreCurrency($this->config->getCurrency());
				$itemModel->setStoreLanguage($this->config->getLanguage());
				$itemModel->setStoreExchangeRate($this->exchangeRate);
				$itemModel->setIsUsaPlugin($this->config->getIsUsaPlugin());
				$itemModel->setSwisscartImageCount($this->swisscartImageCount);
				$itemModel->setSwisscartVersion($this->swisscartVersion);
				$this->addItemModel($itemModel->generateData());
			}
		}
	}

	protected function createItemsCsv()
	{
		$this->itemModel->setLanguageId($this->langID);
		$this->itemModel->setTaxZone($this->taxZone);
		$this->itemModel->setIsUsaPlugin($this->config->getIsUsaPlugin());
		$this->itemModel->setStoreExchangeRate($this->exchangeRate);
		$sqlQuery = $this->itemModel->buildItemQuery(
			$this->exportLimit, $this->exportOffset, $_REQUEST['item_numbers']
		);
		$this->buildCsv($sqlQuery, 'item', array('buildItemOptions'));
	}

	/**
	 * Fill the item_number field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportItemNumber($aItem, $aArticle)
	{
		$aItem["item_number"] = $aArticle["products_id"];

		return $aItem;
	}

	/**
	 * Fill the item_number field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportItemNumberPublic($aItem, $aArticle)
	{
		$aItem["item_number_public"] = $aArticle["products_model"];

		return $aItem;
	}

	/**
	 * Fill the item_name field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportItemName($aItem, $aArticle)
	{
		$aItem["item_name"] = $this->itemModel->buildName($aArticle['products_name']);

		return $aItem;
	}

	/**
	 * Fill the unit_amount field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 *
	 */
	protected function itemExportUnitAmount($aItem, $aArticle)
	{
		$aItem["unit_amount"] = $this->getPrice($aArticle, $aArticle["tax_rate"], $this->exchangeRate);

		return $aItem;
	}

	/**
	 * Fill the currency field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportCurrency($aItem, $aArticle)
	{
		$aItem["currency"] = $this->currency;

		return $aItem;
	}

	/**
	 * Fill the tax_percent field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 *
	 */
	protected function itemExportTaxPercent($aItem, $aArticle)
	{
		$aItem["tax_percent"] = $this->itemModel->buildTaxRate($aArticle["tax_rate"]);

		return $aItem;
	}

	/**
	 * Fill the description field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportDescription($aItem, $aArticle)
	{
		$aItem["description"] = $this->itemModel->buildDescription($aArticle["products_description"]);

		return $aItem;
	}

	/**
	 * Fill the urls_images field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportUrlsImages($aItem, $aArticle)
	{
		$aItem['urls_images'] = implode('||', $this->itemModel->getProductImages($aArticle));

		return $aItem;
	}

	/**
	 * Fill the categories field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportCategories($aItem, $aArticle)
	{
		$aItem["categories"] = $this->getCategories($aArticle);

		return $aItem;
	}

	/**
	 * Takes all categories where the item is mapped to and creates the paths with the actual category-names set
	 *
	 * @param array $aArticle
	 *
	 * @throws ShopgateLibraryException
	 * @return string
	 */
	protected function getCategories($aArticle)
	{
		$categories = "";

		// Make optimized queries possible (get every name only once)
		$categoryIdToNameMapping = array();

		$categoryIds   = $this->getCategoryNumbers($aArticle, false);
		$categoryPaths = array();
		foreach ($categoryIds as $categoryId) {
			$categoryPath = $this->getCategoryPath($categoryId);
			foreach ($categoryPath as $categoryIdInPath) {
				// Saves a list of all used category ids
				$categoryIdToNameMapping[$categoryIdInPath] = $categoryIdInPath;
			}
			$categoryPaths[] = $categoryPath;
		}

		if (!empty($categoryIdToNameMapping)) {
			// Now get all needed category names
			// -> make a list of needed names
			$categoryIdListString = "";
			foreach ($categoryIdToNameMapping as $categoryId) {
				$categoryIdListString .= ',' . $categoryId;
			}
			// -> skip first comma in that list
			$categoryIdListString = substr($categoryIdListString, 1);
			// -> get the names from database
			$sqlQuery = "SELECT DISTINCT " . "tbl_cd.categories_id, "
						. "tbl_cd.categories_name " . "FROM "
						. TABLE_CATEGORIES_DESCRIPTION . " tbl_cd " . "WHERE "
						. "tbl_cd.categories_id IN (" . $categoryIdListString . ") "
						. "AND tbl_cd.language_id = " . $this->langID;

			$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);
			if (!$queryResult) {
				throw new ShopgateLibraryException(
					ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
					"Shopgate Plugin - Error selecting category description.",
					true
				);
			} else {
				// Map the names
				while ($row = ShopgateWrapper::db_fetch_array($queryResult)) {
					$categoryIdToNameMapping[$row["categories_id"]] = $row["categories_name"];
				}
				if (strtoupper($this->config->getLanguage()) == 'GE' || strtoupper($this->config->getLanguage()) == "DE"
				) {
					$categoryIdToNameMapping[0] = "Hauptkategorie";
				} else {
					$categoryIdToNameMapping[0] = "main category";
				}

				// Now build the category string
				if (!empty($categoryPaths)) {
					foreach ($categoryPaths as $categoryPath) {
						if (!empty($categoryPath)) {
							$categories .= '||';
							foreach ($categoryPath as $categoryIdInPath) {
								// Use the name instead of an id and append the "=>" that will be removed after the actual path is finished
								$categories .= $categoryIdToNameMapping[$categoryIdInPath] . "=>";
							}
							// skip the lately appended "=>" because there is no more id to be appended to the actual path
							$categories = substr($categories, 0, -2);
						}
					}
					// skip the first ||
					$categories = substr($categories, 2);
				}
			}
		}

		return $categories;
	}

	/**
	 * Does a query on all categories that the actual product is put in and returns all category numbers as a single string or as an array (if the second param is set to false)
	 *
	 * @param array   $aArticle
	 * @param boolean $returnString
	 *
	 * @throws ShopgateLibraryException
	 * @return string|array
	 */
	protected function getCategoryNumbers($aArticle, $returnString = true)
	{
		$numbers     = array();
		$sqlQuery    = "SELECT DISTINCT " . "tbl_ptc.categories_id " . "FROM "
					   . TABLE_PRODUCTS_TO_CATEGORIES . " tbl_ptc " . "LEFT JOIN "
					   . TABLE_CATEGORIES
					   . " tbl_c ON tbl_ptc.categories_id = tbl_c.categories_id "
					   . "WHERE tbl_ptc.products_id = '" . $aArticle["products_id"]
					   . "' AND tbl_c.categories_id != 0";
		$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);

		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error selecting category numbers.", true
			);
		} else {
			while ($row = ShopgateWrapper::db_fetch_array($queryResult)) {
				$numbers[] = $row["categories_id"];
			}
		}
		// Check if the call er requested the return-value as string
		if ($returnString) {
			// Return all category numbers
			if (!empty($numbers)) {
				return implode('||', $numbers);
			}

			return '';
		}

		// Return an array as told by $returnString
		return $numbers;
	}

	/**
	 * Takes a category id and creates a path-array from the the top category down to the actual one
	 *
	 * @param integer $categoryId
	 *
	 * @return array
	 * @throws ShopgateLibraryException
	 */
	protected function getCategoryPath($categoryId)
	{
		$categoryPath = array();

		if (!empty($categoryId)) {
			// Get parent
			$sqlQuery    = "SELECT " . "tbl_c.parent_id " . "FROM " . TABLE_CATEGORIES
						   . " tbl_c " . "WHERE " . "tbl_c.categories_id = $categoryId "
						   . "LIMIT 1";
			$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);
			if (!$queryResult) {
				throw new ShopgateLibraryException(
					ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
					"Shopgate Plugin - Error selecting category.", true
				);
			} else {
				$row = ShopgateWrapper::db_fetch_array($queryResult);
				if (!empty($row)) {
					if (($row['parent_id'] == 0) || ($row['parent_id'] == $categoryId)) {
						return array($categoryId); // Ignore the main category
					} else {
						// Get all upper categories
						$categoryPath = $this->getCategoryPath($row['parent_id']);
						// append the actual category afterwards
						$categoryPath[] = $categoryId;
					}
				}
			}
		}

		return $categoryPath;
	}

	/**
	 * Fill the category_numbers field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportCategoryNumbers($aItem, $aArticle)
	{
		$aItem["category_numbers"] = $this->getCategoryNumbers($aArticle);

		return $aItem;
	}

	/**
	 * Fill the is_available field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportIsAvailable($aItem, $aArticle)
	{
		$aItem["is_available"] = intval($this->itemModel->getIsAvailable($aArticle['products_quantity']));

		return $aItem;
	}
	
	/**
	 * Fill the available_text field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportAvailableText($aItem, $aArticle)
	{
		$aItem["available_text"] = $this->itemModel->getAvailableText(
			$aArticle['products_quantity'], $aArticle['products_date_available']
		);

		return $aItem;
	}

	/**
	 * Fill the manufacturer field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportManufacturer($aItem, $aArticle)
	{
		$aItem["manufacturer"] = $aArticle["manufacturers_name"];

		return $aItem;
	}

	/**
	 * Fill the url_deeplink field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportUrlDeeplink($aItem, $aArticle)
	{
		$aItem["url_deeplink"] = $this->itemModel->buildDeeplink($aArticle['products_id']);

		return $aItem;
	}

	/**
	 * Fill the old_unit_amount field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 *
	 */
	protected function itemExportOldUnitAmount($aItem, $aArticle)
	{
		$aItem["old_unit_amount"] = $this->getOldPrice($aArticle, $aArticle["tax_rate"], $this->exchangeRate);

		return $aItem;
	}

	/**
	 * Checks if there is a special price and returns the non-special price with tax, if given and adds up exchange if given
	 *
	 * @param array $aArticle
	 * @param null  $taxPercent
	 * @param int   $exchangeValue
	 *
	 * @return float|string empty string if there is no special price or a float value with the old price [with tax and exchange]
	 */
	protected function getOldPrice($aArticle, $taxPercent = null, $exchangeValue = 1) {
		if ($this->isSpecialPrice($aArticle)) {
			$price = $aArticle["products_price"];
		} else {
			$price = "";
		}

		if (!empty($price)) {
			if (!empty($taxPercent)) {
				$price *= 1 + ($taxPercent / 100);
			}
			if (!empty($exchangeValue)) {
				$price *= $exchangeValue;
			}

			$price = $this->formatPriceNumber($price, 2);
		}

		return $price;
	}

	/**
	 * Fill the use_stock field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportUseStock($aItem, $aArticle)
	{
		$aItem["use_stock"] = $this->itemModel->buildUseStock();

		return $aItem;
	}

	/**
	 * Fill the stock_quantity field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportStockQuantity($aItem, $aArticle)
	{
		$aItem["stock_quantity"] = $aArticle['products_quantity'];

		return $aItem;
	}

	/**
	 * Fill the last_update field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportLastUpdate($aItem, $aArticle)
	{
		$aItem["last_update"] = $this->itemModel->buildLastUpdate($aArticle['products_last_modified']);
		
		return $aItem;
	}

	/**
	 * Fill the sort_order field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportSortOrder($aItem, $aArticle)
	{
		$aItem["sort_order"] = $aArticle['products_id'];

		return $aItem;
	}

	/**
	 * Fill the is_highlight field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportIsHighlight($aItem, $aArticle)
	{
		$aItem["is_highlight"] = !empty($aArticle['feature_status']) ? 1 : 0;

		return $aItem;
	}

	/**
	 * Fill the is_highlight field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportHighlightOrderIndex($aItem, $aArticle)
	{
		if (!empty($aArticle['feature_status'])) {
			if (empty($this->itemHighlightOrderIndex)) {
				$this->itemHighlightOrderIndex = 0;
			}
			$aItem["highlight_order_index"] = $this->itemHighlightOrderIndex++;
		} else {
			$aItem["highlight_order_index"] = 0;
		}

		return $aItem;
	}

	/**
	 * Fill the marketplace field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportMarketplace($aItem, $aArticle)
	{
		$aItem["marketplace"] = 1;

		return $aItem;
	}

	/**
	 * Fill the weight field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportWeight($aItem, $aArticle)
	{
		$aItem["weight"] = $aArticle["products_weight"] * 1000;

		return $aItem;
	}

	/**
	 * Takes a list of items and adds option/option_value pairs depending on the options saved for the product
	 *
	 * @param array $itemList
	 * @param array $aArticle
	 *
	 * @return array
	 */
	protected function buildItemOptions($itemList, $aArticle)
	{
		// Accounts only for the first entry (there should also never be passed more than a single item)
		$sgItem = $itemList[0];

		// Get all possible options for the current item
		$sqlQuery      = "SELECT\n" 
						 . "\ttbl_po.products_options_id,\n" 
						 . "\ttbl_po.products_options_name,\n" 
						 . "\ttbl_pov.products_options_values_id,\n"
						 . "\ttbl_pov.products_options_values_name,\n"
						 . "\ttbl_pa.price_prefix,\n" . "\ttbl_pa.options_values_price\n"
						 . "FROM " . TABLE_PRODUCTS_ATTRIBUTES . " AS tbl_pa\n"
						 . "\tLEFT JOIN " . TABLE_PRODUCTS_OPTIONS
						 . " AS tbl_po ON (tbl_pa.options_id = tbl_po.products_options_id)\n"
						 . "\tLEFT JOIN " . TABLE_PRODUCTS_OPTIONS_VALUES
						 . " AS tbl_pov ON (tbl_pa.options_values_id = tbl_pov.products_options_values_id)\n"
						 . "WHERE\n" . "\ttbl_po.language_id = " . ((int)$this->langID) . "\n"
						 . "\t\tAND\n" . "\ttbl_pov.language_id = $this->langID\n"
						 . "\t\tAND\n" . "\ttbl_pa.products_id = $aArticle[products_id];\n";
		
		$queryResult   = ShopgateWrapper::db_query_print_on_err($sqlQuery);
		$options       = array();
		$optionNValues = array();
		while ($row = ShopgateWrapper::db_fetch_array($queryResult)) {
			$optionId             = $row['products_options_id'];
			$optionIdentifierName = $optionId . '=' . $row['products_options_name'];
			if (!in_array($optionIdentifierName, $options)) {
				$options[]                            = $optionIdentifierName;
				$optionNValues[$optionIdentifierName] = array();
			}
			$row['options_values_price'] = trim($row['options_values_price']);
			$prefixModificator           = $row['price_prefix'] == '+' ? 1 : -1;

			// Add taxes and calculate to integer value!
			$valueId      = $row['products_options_values_id'];
			$valueName    = $row['products_options_values_name'];
			$optionsPrice = $this->itemModel->getOptionsValuesPrice(
				$prefixModificator, $row['options_values_price'],
				$aArticle['tax_rate']
			) * 100;

			$optionNValues[$optionId . '=' . $row['products_options_name']][]
				= $valueId . '=' . $valueName . '=>' . $optionsPrice;
		}

		if (!empty($options)) {
			// create a string with valid values for each available option
			foreach ($optionNValues as $optionName => $optionValues) {
				$optionNValues[$optionName] = implode('||', $optionValues);
			}

			// build options for the current product
			foreach ($options as $optionIndex => $optionName) {
				// only options_1 to options_10 allowed
				$optionNumber = $optionIndex + 1;
				if ($optionNumber > 10) {
					break;
				}
				$sgItem['option_' . $optionNumber]             = $optionName;
				$sgItem['option_' . $optionNumber . '_values'] = $optionNValues[$optionName];
			}

			$sgItem['has_options'] = 1;
		}

		$itemList[0] = $sgItem;

		return $itemList;
	}

	###########################################################################
	## CATEGORY EXPORT METHODS                                               ##
	###########################################################################

	/**
	 * @param int|null $limit
	 * @param int|null $offset
	 * @param array    $uids
	 *
	 * @throws ShopgateLibraryException
	 */
	protected function createCategories($limit = null, $offset = null, array $uids = array())
	{
		$sqlQuery         = $this->categoryModel->buildCategoryQuery($this->langID, $limit, $offset, $uids);
		$queryResult      = ShopgateWrapper::db_query_print_on_err($sqlQuery);

		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error selecting Categories.", true
			);
		} else {
			while ($row = ShopgateWrapper::db_fetch_array($queryResult)) {
				$categoryModel = new ShopgateCategoryXmlModel();
				$categoryModel->setItem($row);
				$this->addCategoryModel($categoryModel->generateData());
			}
		}
	}

	/**
	 * @throws ShopgateLibraryException
	 */
	protected function createCategoriesCsv()
	{
		$sqlQuery = $this->categoryModel->buildCategoryQuery($this->langID);
		$this->buildCsv($sqlQuery, 'category');
	}

	/**
	 * Fill the category_number field in the given array
	 *
	 * @param $aCategory
	 * @param $aCategoryRow
	 *
	 * @return array
	 */
	protected function categoryExportCategoryNumber($aCategory, $aCategoryRow)
	{
		$aCategory["category_number"] = $aCategoryRow["categories_id"];

		return $aCategory;
	}

	/**
	 * Fill the category_name field in the given array
	 *
	 * @param $aCategory
	 * @param $aCategoryRow
	 *
	 * @return array
	 */
	protected function categoryExportCategoryName($aCategory, $aCategoryRow)
	{
		$aCategory["category_name"] = $aCategoryRow["categories_name"];

		return $aCategory;
	}

	/**
	 * Fill the parent_id field in the given array
	 *
	 * @param $aCategory
	 * @param $aCategoryRow
	 *
	 * @return array
	 */
	protected function categoryExportParentId($aCategory, $aCategoryRow)
	{
		$aCategory["parent_id"] = $this->categoryModel->buildParentUid(
			$aCategoryRow["parent_id"], $aCategoryRow['categories_id']
		);
		
		return $aCategory;
	}

	/**
	 * Fill the url_image field in the given array
	 *
	 * @param $aCategory
	 * @param $aCategoryRow
	 *
	 * @return array
	 */
	protected function categoryExportUrlImage($aCategory, $aCategoryRow)
	{
		$aCategory["url_image"] = $this->categoryModel->buildCategoryImageUrl($aCategoryRow["categories_image"]);

		return $aCategory;
	}

	/**
	 * Fill the order_index field in the given array
	 *
	 * @param $aCategory
	 * @param $aCategoryRow
	 *
	 * @return array
	 */
	protected function categoryExportOrderIndex($aCategory, $aCategoryRow)
	{
		$aCategory["order_index"] = $this->categoryModel->buildSortOrder($aCategoryRow["sort_order"]);

		return $aCategory;
	}

	/**
	 * Fill the is_active field in the given array
	 *
	 * @param $aCategory
	 * @param $aCategoryRow
	 *
	 * @return array
	 */
	protected function categoryExportIsActive($aCategory, $aCategoryRow)
	{
		// There is no "active" flag for categories inside the shopping-system, so just set it active
		$aCategory["is_active"] = 1;

		return $aCategory;
	}

	/**
	 * Fill the url_deeplink field in the given array
	 *
	 * @param $aCategory
	 * @param $aCategoryRow
	 *
	 * @return array
	 */
	protected function categoryExportUrlDeeplink($aCategory, $aCategoryRow)
	{
		$aCategory["url_deeplink"] = $this->categoryModel->buildDeeplink($aCategoryRow["categories_id"]);

		return $aCategory;
	}

	###########################################################################
	## REVIEW EXPORT METHODS                                                 ##
	###########################################################################

	protected function createReviewsCsv()
	{
		if (defined('PROJECT_VERSION') && strpos(PROJECT_VERSION, '2.2-LC')) {
			return;
		}
		$sqlQuery = $this->reviewModel->buildReviewQuery();
		$this->buildCsv($sqlQuery, 'review');
	}

	/**
	 * @param int|null   $limit
	 * @param int|null   $offset
	 * @param array|null $uids
	 *
	 * @throws ShopgateLibraryException
	 */
	protected function createReviews($limit = null, $offset = null, array $uids = array())
	{
		if (defined('PROJECT_VERSION') && strpos(PROJECT_VERSION, '2.2-LC')) {
			return;
		}
		$sqlQuery    = $this->reviewModel->buildReviewQuery($limit, $offset, $uids);
		$queryResult = ShopgateWrapper::db_query_print_on_err($sqlQuery);

		if (!$queryResult) {
			throw new ShopgateLibraryException(
				ShopgateLibraryException::PLUGIN_DATABASE_ERROR,
				"Shopgate Plugin - Error selecting Reviews.", true
			);
		} else {
			while ($row = ShopgateWrapper::db_fetch_array($queryResult)) {
				$reviewModel = new ShopgateReviewXmlModel();;
				$reviewModel->setItem($row);
				$this->addReviewModel($reviewModel->generateData());
			}
		}
	}

	/**
	 * Fill the item_number field in the given array
	 *
	 * @param $aReview
	 * @param $aReviewRow
	 *
	 * @return array
	 */
	protected function reviewExportItemNumber($aReview, $aReviewRow)
	{
		$aReview["item_number"] = $aReviewRow["products_id"];

		return $aReview;
	}

	/**
	 * Fill the update_review_id field in the given array
	 *
	 * @param $aReview
	 * @param $aReviewRow
	 *
	 * @return array
	 */
	protected function reviewExportUpdateReviewId($aReview, $aReviewRow)
	{
		$aReview["update_review_id"] = $aReviewRow["reviews_id"];

		return $aReview;
	}

	/**
	 * Fill the score field in the given array
	 *
	 * @param $aReview
	 * @param $aReviewRow
	 *
	 * @return array
	 */
	protected function reviewExportScore($aReview, $aReviewRow)
	{
		$aReview["score"] = $this->reviewModel->buildScore($aReviewRow["reviews_rating"]);

		return $aReview;
	}

	/**
	 * Fill the name field in the given array
	 *
	 * @param $aReview
	 * @param $aReviewRow
	 *
	 * @return array
	 */
	protected function reviewExportName($aReview, $aReviewRow)
	{
		$aReview["name"] = $aReviewRow["customers_name"];

		return $aReview;
	}

	/**
	 * Fill the date field in the given array
	 *
	 * @param $aReview
	 * @param $aReviewRow
	 *
	 * @return array
	 */
	protected function reviewExportDate($aReview, $aReviewRow)
	{
		$aReview["date"] = $this->reviewModel->buildDate($aReviewRow['date_added']);

		return $aReview;
	}

	/**
	 * Fill the title field in the given array
	 *
	 * @param $aReview
	 * @param $aReviewRow
	 *
	 * @return array
	 */
	protected function reviewExportTitle($aReview, $aReviewRow)
	{
		$aReview["title"] =  $this->reviewModel->buildTitle($aReviewRow['reviews_text']);

		return $aReview;
	}

	/**
	 * Fill the text field in the given array
	 *
	 * @param $aReview
	 * @param $aReviewRow
	 *
	 * @return array
	 */
	protected function reviewExportText($aReview, $aReviewRow)
	{
		$aReview["text"] = $aReviewRow["reviews_text"];

		return $aReview;
	}
}

class ShopgatePluginOsCommerceUsa extends ShopgatePluginOsCommerce
{
	public function startup()
	{
		parent::startup();
		$this->useTaxClasses();
	}
	
	/**
	 * Fill the weight field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportWeight($aItem, $aArticle)
	{
		$aItem["weight"] = $aArticle["products_weight"] * self::CONVERT_POUNDS_TO_GRAM_FACTOR;

		return $aItem;
	}

	/**
	 * Fill the unit_amount_net field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return array
	 */
	protected function itemExportUnitAmountNet($aItem, $aArticle)
	{
		$aItem['unit_amount_net'] = $this->getPrice($aArticle, null, $this->exchangeRate);

		return $aItem;
	}

	/**
	 * Fill the old_unit_amount_net field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return mixed
	 */
	protected function itemExportOldUnitAmountNet($aItem, $aArticle)
	{
		$aItem['old_unit_amount_net'] = $this->getOldPrice($aArticle, null, $this->exchangeRate);

		return $aItem;
	}

	/**
	 * Fill the tax_class field in the given array
	 *
	 * @param $aItem
	 * @param $aArticle
	 *
	 * @return mixed
	 */
	protected function itemExportTaxClass($aItem, $aArticle)
	{
		$aItem['tax_class'] = $this->itemModel->buildTaxClass($aArticle);

		return $aItem;
	}
}

class ShopgateMapper
{
	/**
	 * The countries with non-ISO-3166-2 state codes in the shoppingsystem are mapped here.
	 *
	 * @var string[][]
	 */
	protected static $stateCodesByCountryCode
		= array(
			'DE' => array(
				"BW" => "BAW",
				"BY" => "BAY",
				"BE" => "BER",
				"BB" => "BRG",
				"HB" => "BRE",
				"HH" => "HAM",
				"HE" => "HES",
				"MV" => "MEC",
				"NI" => "NDS",
				"NW" => "NRW",
				"RP" => "RHE",
				"SL" => "SAR",
				"SN" => "SAS",
				"ST" => "SAC",
				"SH" => "SCN",
				"TH" => "THE",
			),
			"AT" => array(
				"1" => "BL",
				"2" => "KN",
				"3" => "NO",
				"4" => "OO",
				"5" => "SB",
				"6" => "ST",
				"7" => "TI",
				"8" => "VB",
				"9" => "WI",
			),
			// "CH" => already correct
			// "US" => already correct
		);

	/**
	 * Finds the corresponding Shopgate state code for a given shoppingsystem state code (zone_code).
	 *
	 * @param string $countryCode The code of the country to which the state belongs
	 * @param string $stateCode   The code of the state / zone as found in the default "zones" table of ZenCart
	 *
	 * @return string The state code as defined at Shopgate Wiki
	 *
	 * @throws ShopgateLibraryException if one of the given codes is unknown
	 */
	public static function getShopgateStateCode($countryCode, $stateCode)
	{
		$countryCode = strtoupper($countryCode);
		$stateCode   = strtoupper($stateCode);

		if (!isset(self::$stateCodesByCountryCode[$countryCode])) {
			return $countryCode . '-' . $stateCode;
		}

		$codes = array_flip(self::$stateCodesByCountryCode[$countryCode]);
		if (!isset($codes[$stateCode])) {
			return $countryCode . '-' . $stateCode;
		}

		$returnStateCode = $codes[$stateCode];
		
		return $countryCode . '-' . $returnStateCode;
	}

	/**
	 * Finds the corresponding shoppingsystem state code (zone_code) for a given Shopgate state code
	 *
	 * @param string $shopgateStateCode The Shopgate state code as defined at Shopgate Wiki
	 *
	 * @return string The zone code for xt:Commerce 3
	 *
	 * @throws ShopgateLibraryException if the given code is unknown
	 */
	public static function getShoppingsystemStateCode($shopgateStateCode)
	{
		$splitCodes = null;
		preg_match('/^([A-Z]{2})\-([A-Z]{2})$/', $shopgateStateCode, $splitCodes);

		if (empty($splitCodes) || empty($splitCodes[1]) || empty($splitCodes[2])) {
			return null;
		}

		if (!isset(self::$stateCodesByCountryCode[$splitCodes[1]])
			|| !isset(self::$stateCodesByCountryCode[$splitCodes[1]][$splitCodes[2]])
		) {
			return $splitCodes[2];
		} else {
			return self::$stateCodesByCountryCode[$splitCodes[1]][$splitCodes[2]];
		}
	}
}
